# -*- coding: utf-8 -*-


import os
import time
import json

import pytest
import shutil
import pandas

from academic_tracker.athr_srch_emails_and_reports import create_pubs_by_author_dict, create_project_reports_and_emails, create_project_report
from academic_tracker.athr_srch_emails_and_reports import create_summary_report, build_author_loop, create_collaborators_reports_and_emails
from academic_tracker.athr_srch_emails_and_reports import create_tabular_collaborator_report, create_collaborator_report, create_tabular_summary_report
from academic_tracker.athr_srch_emails_and_reports import create_tabular_project_report, _build_report_rows
from academic_tracker.fileio import load_json, read_text_from_txt


TESTING_DIR = "test_dir"
@pytest.fixture(scope="module", autouse=True)
def test_email_dir():
    save_dir_name = TESTING_DIR
    os.mkdir(save_dir_name)
    time_to_wait = 10
    time_counter = 0
    
    while not os.path.exists(save_dir_name):
        time.sleep(1)
        time_counter += 1
        if time_counter > time_to_wait:
            raise FileNotFoundError(save_dir_name + " was not created within " + str(time_to_wait) + " seconds, so it is assumed that it won't be and something went wrong.")



@pytest.fixture
def publication_dict():
    return load_json(os.path.join("tests", "testing_files", "publication_dict_truncated.json"))


def test_create_pubs_by_author_dict(publication_dict):
    expected = load_json(os.path.join("tests", "testing_files", "pubs_by_author_dict_truncated.json"))
    
    actual = create_pubs_by_author_dict(publication_dict)
    
    assert expected == actual



@pytest.fixture
def config_dict():
    return load_json(os.path.join("tests", "testing_files", "config_truncated.json"))


@pytest.fixture
def authors_by_project_dict():
    return load_json(os.path.join("tests", "testing_files", "authors_by_project_dict_truncated.json"))


def test_create_project_reports_and_emails(publication_dict, config_dict, authors_by_project_dict):
    
    ## Add Travis to a publication so we can test that multiple authors are sent emails.
    publication_dict["https://doi.org/10.1038/s41597-023-02277-x"]["authors"][1]["author_id"] = "Travis Thompson"
    
    expected_emails = load_json(os.path.join("tests", "testing_files", "athr_project_emails.json"))
    del expected_emails["creation_date"]
    
    actual_emails = create_project_reports_and_emails(authors_by_project_dict, publication_dict, config_dict, TESTING_DIR)
    # with open(os.path.join("tests", "testing_files", "athr_project_emails_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_emails, indent=2, sort_keys=True))
    del actual_emails["creation_date"]
    
    dir_contents = os.listdir(TESTING_DIR)
    expected_num_of_project_reports = 5
    actual_num_of_project_reports = len(dir_contents)
    
    assert expected_emails == actual_emails
    assert actual_num_of_project_reports == expected_num_of_project_reports
    
    
@pytest.fixture
def config_dict_tabular():
    return load_json(os.path.join("tests", "testing_files", "config_tabular.json"))


@pytest.fixture
def authors_by_project_dict_tabular():
    return load_json(os.path.join("tests", "testing_files", "authors_by_project_dict_tabular.json"))
    
    
def test_create_project_reports_and_emails_tabular(publication_dict, config_dict_tabular, authors_by_project_dict_tabular):
    ## Add a filename to a project report to test that it will create it correctly.
    authors_by_project_dict_tabular["Core A Administrative Core"]["Hunter Moseley"]["project_report"]["filename"] = "asdf.csv"
    ## Add Travis to a publication so we can test that multiple authors are sent emails.
    publication_dict["https://doi.org/10.1038/s41597-023-02277-x"]["authors"][1]["author_id"] = "Travis Thompson"
            
    expected_emails = load_json(os.path.join("tests", "testing_files", "athr_project_emails_tabular.json"))
    del expected_emails["creation_date"]
    ## attachment for xlsx files is a filepath and it differs between OS's.
    del expected_emails["emails"][1]["attachment"]
    
    actual_emails = create_project_reports_and_emails(authors_by_project_dict_tabular, publication_dict, config_dict_tabular, TESTING_DIR)
    # with open(os.path.join("tests", "testing_files", "athr_project_emails_tabular_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_emails, indent=2, sort_keys=True))
    del actual_emails["creation_date"]
    del actual_emails["emails"][1]["attachment"]
    
    dir_contents = os.listdir(TESTING_DIR)
    expected_num_of_project_reports = 5
    actual_num_of_project_reports = len(dir_contents)
    
    number_of_excel_files = len([name for name in dir_contents if ".xlsx" in name])
    
    assert "asdf.csv" in dir_contents
    assert expected_emails == actual_emails
    assert number_of_excel_files == 1
    assert "no_from_email_report.csv" in dir_contents
    assert actual_num_of_project_reports == expected_num_of_project_reports
        



def test_create_project_report(publication_dict, config_dict, authors_by_project_dict):
    ## Add Travis to a publication so we can test that multiple authors get reported.
    publication_dict["https://doi.org/10.1038/s41597-023-02277-x"]["authors"][1]["author_id"] = "Travis Thompson"
    
    template_string = "<author_first> <author_last>:\n<author_loop><author_first> <author_last>:\n<pub_loop>\t<title> <authors> <grants>\n</pub_loop></author_loop>"
    author_first = "Kelly"
    author_last = "Pennell"
    
    expected_text = read_text_from_txt(os.path.join("tests", "testing_files", "project_report.txt"))
    
    actual_text = create_project_report(publication_dict, config_dict, authors_by_project_dict, "No Authors", template_string, author_first, author_last)
    # with open(os.path.join("tests", "testing_files", "project_report_new.txt"), 'wb') as outFile:
    #     outFile.write(actual_text.encode("utf-8"))
    
    assert expected_text == actual_text



def test_create_summary_report(publication_dict, config_dict, authors_by_project_dict):
    ## Add Travis to a publication so we can test that multiple authors get reported.
    publication_dict["https://doi.org/10.1038/s41597-023-02277-x"]["authors"][1]["author_id"] = "Travis Thompson"
    
    expected_text = read_text_from_txt(os.path.join("tests", "testing_files", "athr_srch_summary_report.txt"))
    
    actual_text = create_summary_report(publication_dict, config_dict, authors_by_project_dict)
    # with open(os.path.join("tests", "testing_files", "athr_srch_summary_report_new.txt"), 'wb') as outFile:
    #     outFile.write(actual_text.encode("utf-8"))
    
    assert expected_text == actual_text



def test_build_author_loop(publication_dict, config_dict, authors_by_project_dict):
    
    template_string = read_text_from_txt(os.path.join("tests", "testing_files", "athr_srch_build_loop_template_string.txt"))
                         
    expected_text = read_text_from_txt(os.path.join("tests", "testing_files", "athr_srch_build_author_loop.txt"))
    
    actual_text = build_author_loop(publication_dict, config_dict, authors_by_project_dict, "No from_email", template_string)
    # with open(os.path.join("tests", "testing_files", "athr_srch_build_author_loop_new.txt"), 'wb') as outFile:
    #     outFile.write(actual_text.encode("utf-8"))
    
    assert expected_text == actual_text
  
        


def test_create_collaborators_reports_and_emails_tabular(publication_dict, config_dict):
    ## Add Travis and Sweta to a publication so we can test that multiple authors get reported.
    publication_dict["https://doi.org/10.1038/s41597-023-02277-x"]["authors"][1]["author_id"] = "Travis Thompson"
    publication_dict["https://doi.org/10.1038/s41597-023-02277-x"]["authors"][0]["author_id"] = "Sweta Ojha"
    ## Add an author that won't be in config_dict to test that line of code.
    publication_dict["https://doi.org/10.1038/s41597-023-02277-x"]["authors"][2]["author_id"] = "asdf"
    
    config_dict["Authors"]["Sweta Ojha"]["collaborator_report"] = {"from_email":"ptth222@uky.edu", 
                                                                    "email_body":"asdf",
                                                                    "email_subject":"asdf",
                                                                    }
    config_dict["Authors"]["Travis Thompson"]["collaborator_report"] = {"from_email":"ptth222@uky.edu",
                                                                        "to_email":"ptth222@uky.edu",
                                                                        "email_body":"asdf",
                                                                        "email_subject":"asdf",
                                                                        "columns":{"Col1":"<pub_author_first>, <pub_author_last>, <pub_author_initials>, <pub_author_affiliations>"},
                                                                        "separator":"\t"
                                                                      }
    config_dict["Authors"]["Hunter Moseley"]["collaborator_report"] = {"from_email":"ptth222@uky.edu",
                                                                        "to_email":"ptth222@uky.edu",
                                                                        "email_body":"asdf",
                                                                        "email_subject":"asdf",
                                                                        "columns":{"Col1":"<pub_author_first>, <pub_author_last>, <pub_author_initials>, <pub_author_affiliations>"},
                                                                        "sort":["Col1"]
                                                                        }
    
    expected_emails = load_json(os.path.join("tests", "testing_files", "collaborator_emails_tabular.json"))
    del expected_emails["creation_date"]
    
    actual_emails = create_collaborators_reports_and_emails(publication_dict, config_dict, TESTING_DIR)
    # with open(os.path.join("tests", "testing_files", "collaborator_emails_tabular_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_emails, indent=2, sort_keys=True))
    del actual_emails["creation_date"]
    
    dir_contents = os.listdir(TESTING_DIR)
    expected_num_of_collaborator_reports = 3
    actual_num_of_collaborator_reports = len(dir_contents)
    
    assert expected_emails == actual_emails
    assert actual_num_of_collaborator_reports == expected_num_of_collaborator_reports
    


def test_create_collaborators_reports_and_emails(publication_dict, config_dict):
    ## Add Travis, Sweta, and Kelly to a publication so we can test that multiple authors get reported.
    publication_dict["https://doi.org/10.1038/s41597-023-02277-x"]["authors"][1]["author_id"] = "Travis Thompson"
    publication_dict["https://doi.org/10.1038/s41597-023-02277-x"]["authors"][0]["author_id"] = "Sweta Ojha"
    publication_dict["https://doi.org/10.1038/s41597-023-02277-x"]["authors"][4]["author_id"] = "Kelly Pennell"
    
    config_dict["Authors"]["Travis Thompson"]["collaborator_report"] = {"template":"<pub_author_loop><pub_author_first>, <pub_author_last>, <pub_author_initials>, <pub_author_affiliations></pub_author_loop>",
                                                                        "from_email":"ptth222@uky.edu",
                                                                        "email_body":"asdf",
                                                                        "email_subject":"asdf"}
    
    config_dict["Authors"]["Sweta Ojha"]["collaborator_report"] = {"template":"<pub_author_loop><pub_author_first>, <pub_author_last>, <pub_author_initials>, <pub_author_affiliations></pub_author_loop>",
                                                                    "from_email":"ptth222@uky.edu",
                                                                    "email_body":"asdf",
                                                                    "email_subject":"asdf",
                                                                    "filename":"name_test.txt"}
    
    config_dict["Authors"]["Kelly Pennell"]["collaborator_report"] = {"from_email":"ptth222@uky.edu",
                                                                      "to_email":"ptth222@uky.edu",
                                                                      "email_body":"asdf",
                                                                      "email_subject":"asdf",
                                                                      "columns":{"Col1":"<pub_author_first>, <pub_author_last>, <pub_author_initials>, <pub_author_affiliations>"},
                                                                      "file_format":"xlsx"}
    
    config_dict["Authors"]["Hunter Moseley"]["collaborator_report"] = {"from_email":"ptth222@uky.edu",
                                                                        "to_email":"ptth222@uky.edu",
                                                                        "email_body":"asdf",
                                                                        "email_subject":"asdf",
                                                                        "columns":{"Col1":"<pub_author_first>, <pub_author_last>, <pub_author_initials>, <pub_author_affiliations>"},
                                                                        "sort":["Col1"],
                                                                        "filename":"name_test.csv"}
    
    expected_emails = load_json(os.path.join("tests", "testing_files", "collaborator_emails.json"))
    del expected_emails["creation_date"]
    del expected_emails["emails"][3]["attachment"]
    
    actual_emails = create_collaborators_reports_and_emails(publication_dict, config_dict, TESTING_DIR)
    # with open(os.path.join("tests", "testing_files", "collaborator_emails_new.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_emails, indent=2, sort_keys=True))
    del actual_emails["creation_date"]
    del actual_emails["emails"][3]["attachment"]
    
    dir_contents = os.listdir(TESTING_DIR)
    expected_num_of_collaborator_reports = 4
    actual_num_of_collaborator_reports = len(dir_contents)
    
    number_of_excel_files = len([name for name in dir_contents if ".xlsx" in name])
    
    assert expected_emails == actual_emails
    assert "name_test.txt" in dir_contents
    assert "name_test.csv" in dir_contents
    assert number_of_excel_files == 1
    assert actual_num_of_collaborator_reports == expected_num_of_collaborator_reports
    
    


@pytest.mark.parametrize("config_dict_collab, filename, file_format", [
        
        ({"Authors":{"Hunter Moseley":{"collaborator_report":{"columns":{"Col1":"<pub_author_first>, <pub_author_last>, <pub_author_initials>, <pub_author_affiliations>", "Col2":"asdf"},
                                                            "separator":"\t",
                                                            "sort":["Col1"],
                                                            "column_order":["Col2", "Col1"]}}}}, "test_name.csv", "csv"),
        ({"Authors":{"Hunter Moseley":{"collaborator_report":{"columns":{"Col1":"<pub_author_first>, <pub_author_last>, <pub_author_initials>, <pub_author_affiliations>", "Col2":"asdf"},
                                                            "column_order":["Col1", "Col2"]}}}}, "test_name.csv", "csv"),
        ])   

def test_create_tabular_collaborator_report(publication_dict, config_dict_collab, filename, file_format):
    
    if "separator" in config_dict_collab["Authors"]["Hunter Moseley"]["collaborator_report"]:
        sep = config_dict_collab["Authors"]["Hunter Moseley"]["collaborator_report"]["separator"]
    else:
        sep = ","
        
    if "sort" in config_dict_collab["Authors"]["Hunter Moseley"]["collaborator_report"]:
        column_one = 'Atta M, Nawabi, AM, Division of Transplant and Hepatobiliary, Department of Surgery, The University of Kansas Medical Center, Kansas City, Kansas, USA.'
    else:
        column_one = 'Zachary A, Kipp, ZA, Department of Pharmacology and Nutritional Sciences, University of Kentucky College of Medicine, Lexington, Kentucky, USA.'
    
    create_tabular_collaborator_report(publication_dict, config_dict_collab, "Hunter Moseley", ["https://doi.org/10.1002/hep.32467"], filename, file_format, TESTING_DIR)
    
    dir_contents = os.listdir(TESTING_DIR)
    
    assert filename in dir_contents
    
    df = pandas.read_csv(os.path.join(TESTING_DIR, filename), sep=sep)
    report_text = read_text_from_txt(os.path.join(TESTING_DIR, filename))
    # df.to_csv(os.path.join("tests", "testing_files", "aaa.csv"),sep=sep, index=False, lineterminator="\n")
    assert df.to_csv(sep=sep, index=False, lineterminator="\n") == report_text
    assert list(df.columns) == config_dict_collab["Authors"]["Hunter Moseley"]["collaborator_report"]["column_order"]
    assert df.iloc[1].loc["Col1"] == column_one
    assert df.iloc[0].loc["Col2"] == "asdf"
    
    

def test_create_tabular_collaborator_report_defaults(publication_dict):
    
    config_dict = {"Authors":{"Hunter Moseley":{"collaborator_report":{}}}}
    
    create_tabular_collaborator_report(publication_dict, config_dict, "Hunter Moseley", ["https://doi.org/10.1002/hep.32467"], "test_name.csv", "csv", TESTING_DIR)
    
    dir_contents = os.listdir(TESTING_DIR)
    
    assert "test_name.csv" in dir_contents
    
    df = pandas.read_csv(os.path.join(TESTING_DIR, "test_name.csv"), sep=",")
    report_text = read_text_from_txt(os.path.join(TESTING_DIR, "test_name.csv"))
    assert df.to_csv(sep=",", index=False, lineterminator="\n") == report_text
    assert list(df.columns) == ["Name", "Affiliations"]
    assert df.iloc[0].loc["Name"] == 'Alganem, Khaled'
    assert df.iloc[0].loc["Affiliations"] == 'Department of Neurosciences, University of Toledo College of Medicine and Life Sciences, Toledo, Ohio, USA.'
    

def test_create_tabular_collaborator_report_excel(publication_dict):
    
    config_dict = {"Authors":{"Hunter Moseley":{"collaborator_report":{}}}}
    
    create_tabular_collaborator_report(publication_dict, config_dict, "Hunter Moseley", ["https://doi.org/10.1002/hep.32467"], "test_name.csv", "xlsx", TESTING_DIR)
    
    dir_contents = os.listdir(TESTING_DIR)
    
    assert "test_name.csv.xlsx" in dir_contents
    
    df = pandas.read_excel(os.path.join(TESTING_DIR, "test_name.csv.xlsx"))
    assert list(df.columns) == ["Name", "Affiliations"]
    assert df.iloc[0].loc["Name"] == 'Alganem, Khaled'
    assert df.iloc[0].loc["Affiliations"] == 'Department of Neurosciences, University of Toledo College of Medicine and Life Sciences, Toledo, Ohio, USA.'
    
    
def test_create_tabular_collaborator_report_empty():
    
    config_dict = {"Authors":{"Hunter Moseley":{"collaborator_report":{}}}}
    publication_dict = {"https://doi.org/10.1002/hep.32467":{"authors":[]}}
    
    report, filename = create_tabular_collaborator_report(publication_dict, config_dict, "Hunter Moseley", ["https://doi.org/10.1002/hep.32467"], "test_name.csv", "xlsx", TESTING_DIR)
    
    assert report == ""
    
    
    
def test_create_collaborator_report(publication_dict):
    
    template = "<pub_author_loop><pub_author_first>, <pub_author_last>, <pub_author_initials>, <pub_author_affiliations></pub_author_loop>"
    
    expected_report = read_text_from_txt(os.path.join("tests", "testing_files", "collaborator_report2.txt"))
    
    actual_report = create_collaborator_report(publication_dict, template, "Hunter Moseley", ["https://doi.org/10.1002/hep.32467"], "test_name.txt", TESTING_DIR)
    # with open(os.path.join("tests", "testing_files", "collaborator_report2_new.txt"), 'wb') as outFile:
    #     outFile.write(actual_report.encode("utf-8"))
    
    dir_contents = os.listdir(TESTING_DIR)
    
    assert "test_name.txt" in dir_contents
    assert expected_report == actual_report
    
    



def test_create_tabular_summary_report_no_pub_keywords(publication_dict, config_dict, authors_by_project_dict):
    
    config_dict["summary_report"] = {"columns":{"Col1":"<project_name>", "Col2":"<author_first>"},
                                      "filename":"test_name.csv",
                                      "file_format":"csv",
                                      "separator":"\t",
                                      "sort":["Col2"],
                                      "column_order":["Col2", "Col1"]}
    
    create_tabular_summary_report(publication_dict, config_dict, authors_by_project_dict, TESTING_DIR)
    
    dir_contents = os.listdir(TESTING_DIR)
    
    assert "test_name.csv" in dir_contents
    
    df = pandas.read_csv(os.path.join(TESTING_DIR, "test_name.csv"), sep="\t")
    report_text = read_text_from_txt(os.path.join(TESTING_DIR, "test_name.csv"))
    # df.to_csv(os.path.join("tests", "testing_files", "aaa.csv"), sep="\t", index=False, lineterminator="\n")
    assert df.to_csv(sep="\t", index=False, lineterminator="\n") == report_text
    assert list(df.columns) == config_dict["summary_report"]["column_order"]
    assert df.iloc[1].loc["Col1"] == "Project 1"
    assert df.iloc[0].loc["Col2"] == "Hunter"
    
    
def test_create_tabular_summary_report_defaults(publication_dict, config_dict, authors_by_project_dict):
    
    config_dict["summary_report"] = {"columns":{"Col1":"<project_name>", "Col2":"<author_first>", "Col3":"<title>"}}
    
    create_tabular_summary_report(publication_dict, config_dict, authors_by_project_dict, TESTING_DIR)
    
    dir_contents = os.listdir(TESTING_DIR)
    
    assert "summary_report.csv" in dir_contents
    
    df = pandas.read_csv(os.path.join(TESTING_DIR, "summary_report.csv"), sep=",")
    # df.to_csv(os.path.join("tests", "testing_files", "aaa.csv"), sep=",", index=False, lineterminator="\n")
    report_text = read_text_from_txt(os.path.join(TESTING_DIR, "summary_report.csv"))
    assert df.to_csv(sep=",", index=False, lineterminator="\n") == report_text
    assert list(df.columns) == list(config_dict["summary_report"]["columns"].keys())
    assert df.iloc[1].loc["Col1"] == 'Core A Administrative Core'
    assert df.iloc[0].loc["Col2"] == "Hunter"
    assert df.iloc[0].loc["Col3"] == 'Plk1 phosphorylation of PHGDH to regulate serine metabolism'
    

def test_create_tabular_summary_report_excel(publication_dict, config_dict, authors_by_project_dict):
    
    config_dict["summary_report"] = {"columns":{"Col1":"<project_name>", "Col2":"<author_first>", "Col3":"<title>", "Col4":"<pub_author_first>"},
                                      "filename":"summary_report.csv",
                                      "file_format":"xlsx"}
    
    create_tabular_summary_report(publication_dict, config_dict, authors_by_project_dict, TESTING_DIR)
    
    dir_contents = os.listdir(TESTING_DIR)
    
    assert "summary_report.csv.xlsx" in dir_contents
    
    df = pandas.read_excel(os.path.join(TESTING_DIR, "summary_report.csv.xlsx"))
    # df.to_csv(os.path.join("tests", "testing_files", "aaa.csv"), sep=",", index=False, lineterminator="\n")
    assert list(df.columns) == list(config_dict["summary_report"]["columns"].keys())
    assert df.iloc[1].loc["Col1"] == 'Core A Administrative Core'
    assert df.iloc[0].loc["Col2"] == "Hunter"
    assert df.iloc[0].loc["Col3"] == 'Plk1 phosphorylation of PHGDH to regulate serine metabolism'
    assert df.iloc[0].loc["Col4"] == "Hunter"





@pytest.fixture
def pubs_by_author_dict():
    return load_json(os.path.join("tests", "testing_files", "pubs_by_author_dict_truncated.json"))


def test_build_report_rows_no_authors(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict):
    "There should never be a publication without an author, but the code is there just in case."
    publication_dict["https://doi.org/10.1002/hep.32467"]["authors"] = []
    
    row_template = {"Col1":"<project_name>", "Col2":"<author_first>", "Col3":"<title>", "Col4":"<pub_author_first>"}
    
    rows = _build_report_rows(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict, row_template, "Project 1")
    assert rows[1]["Col4"] == "None"
    

def test_build_report_rows_only_reference_keywords(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict):
    "Need to test a template with reference keywords and no author keywords."
    publication_dict["https://doi.org/10.1002/hep.32467"]["authors"] = []
    
    row_template = {"Col1":"<project_name>", "Col2":"<author_first>", "Col3":"<title>", "Col4":"<reference_citation>"}
    
    rows = _build_report_rows(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict, row_template, "Project 1")
    assert rows[0]["Col4"] == "None"
    assert len(rows) == 81
    assert rows[1]["Col4"] == "Roehlen N, Crouchet E, Baumert TF. Liver fibrosis: mechanistic concepts and therapeutic perspectives. Cells. 2020;9:875."


def test_create_tabular_project_report_no_pub_keywords(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict):
    
    report_attributes = {"columns":{"Col1":"<project_name>", "Col2":"<author_first>"},
                          "filename":"test_name.csv",
                          "file_format":"csv",
                          "separator":"\t",
                          "sort":["Col2"],
                          "column_order":["Col2", "Col1"]}
    
    project_name = "Project 1"
    
    filename = "project_report.csv"
    
    create_tabular_project_report(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict, project_name, report_attributes, TESTING_DIR, filename)
    
    dir_contents = os.listdir(TESTING_DIR)
    
    assert filename in dir_contents
    
    df = pandas.read_csv(os.path.join(TESTING_DIR, filename), sep="\t")
    # df.to_csv(os.path.join("tests", "testing_files", "aaa.csv"), sep="\t", index=False, lineterminator="\n")
    report_text = read_text_from_txt(os.path.join(TESTING_DIR, filename))
    assert df.to_csv(sep="\t", index=False, lineterminator="\n") == report_text
    assert list(df.columns) == report_attributes["column_order"]
    assert df.iloc[0].loc["Col1"] == "Project 1"
    assert df.iloc[0].loc["Col2"] == "Hunter"
    
    

def test_create_tabular_project_report_defaults(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict):
    
    report_attributes = {"columns":{"Col1":"<project_name>", "Col2":"<author_first>", "Col3":"<title>"}}
    
    project_name = "Project 1"
    
    filename = "project_report.csv"
    
    create_tabular_project_report(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict, project_name, report_attributes, TESTING_DIR, filename)
    
    dir_contents = os.listdir(TESTING_DIR)
    
    assert "project_report.csv" in dir_contents
    
    df = pandas.read_csv(os.path.join(TESTING_DIR, "project_report.csv"), sep=",")
    # df.to_csv(os.path.join("tests", "testing_files", "aaa.csv"), sep=",", index=False, lineterminator="\n")
    report_text = read_text_from_txt(os.path.join(TESTING_DIR, "project_report.csv"))
    assert df.to_csv(sep=",", index=False, lineterminator="\n") == report_text
    assert list(df.columns) == list(report_attributes["columns"].keys())
    assert df.iloc[1].loc["Col1"] == 'Project 1'
    assert df.iloc[0].loc["Col2"] == "Hunter"
    assert df.iloc[0].loc["Col3"] == 'Plk1 phosphorylation of PHGDH to regulate serine metabolism'
    

def test_create_tabular_project_report_excel(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict):
    
    report_attributes = {"columns":{"Col1":"<project_name>", 
                                    "Col2":"<author_first>", 
                                    "Col3":"<title>", 
                                    "Col4":"<pub_author_first>", 
                                    "Col5":"<reference_citation>"},
                          "filename":"project_report.csv",
                          "file_format":"xlsx"}
    
    project_name = "No Authors"
    
    filename = "project_report.csv"
    
    create_tabular_project_report(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict, project_name, report_attributes, TESTING_DIR, filename)
    
    dir_contents = os.listdir(TESTING_DIR)
    
    assert "project_report.csv.xlsx" in dir_contents
    
    df = pandas.read_excel(os.path.join(TESTING_DIR, "project_report.csv.xlsx"), keep_default_na=False)
    # df.to_csv(os.path.join("tests", "testing_files", "aaa.csv"), sep="\t", index=False, lineterminator="\n")
    assert list(df.columns) == list(report_attributes["columns"].keys())
    assert df.iloc[1].loc["Col1"] == 'No Authors'
    assert df.iloc[0].loc["Col2"] == "Hunter"
    assert df.iloc[0].loc["Col3"] == 'Plk1 phosphorylation of PHGDH to regulate serine metabolism'
    assert df.iloc[0].loc["Col4"] == "Hunter"
    assert df.iloc[0].loc["Col5"] == "None"
    assert df.iloc[1].loc["Col3"] == 'Hepatic kinome atlas: An in-depth identification of kinase pathways in liver fibrosis of humans and rodents.'
    assert df.iloc[811].loc["Col3"] == 'Identifying and sharing per-and polyfluoroalkyl substances hot-spot areas and exposures in drinking water.'
    assert df.iloc[1].loc["Col5"] == 'Roehlen N, Crouchet E, Baumert TF. Liver fibrosis: mechanistic concepts and therapeutic perspectives. Cells. 2020;9:875.'
    assert df.iloc[46].loc["Col4"] == "Zachary A"




@pytest.fixture(autouse=True)
def test_dir_reset(request):
    def clean_test_dir():
        if os.path.exists(TESTING_DIR):
            for test_file in os.listdir(TESTING_DIR):
                os.remove(os.path.join(TESTING_DIR, test_file))
    request.addfinalizer(clean_test_dir)


@pytest.fixture(scope="module", autouse=True)
def cleanup(request):
    """Cleanup a testing directory once we are finished."""
    def remove_test_dir():
        if os.path.exists(TESTING_DIR):
            shutil.rmtree(TESTING_DIR)
    request.addfinalizer(remove_test_dir)




# -*- coding: utf-8 -*-

import os
import requests
import copy
import re
import xml.etree.ElementTree as ET
import json

import pytest
import shutil
import pymed

from academic_tracker.ref_srch_modularized import input_reading_and_checking, build_publication_dict, save_and_send_reports_and_emails
from academic_tracker.fileio import load_json, read_text_from_txt
from academic_tracker.ref_srch_webio import search_references_on_source


@pytest.fixture(autouse=True)
def disable_network_calls(monkeypatch):
    def stunted_get():
        raise RuntimeError("Network access not allowed during testing!")
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: stunted_get())



@pytest.fixture
def config_dict():
    return load_json(os.path.join("tests", "testing_files", "config_truncated.json"))

@pytest.fixture
def config_dict_Hunter_only():
    return load_json(os.path.join("tests", "testing_files", "config_Hunter_only.json"))

@pytest.fixture
def tokenized_citations():
    return load_json(os.path.join("tests", "testing_files", "tokenized_ref_test.json"))

@pytest.fixture
def original_queries():
    query_json = load_json(os.path.join("tests", "testing_files", "all_queries_ref.json"))
    ## Convert PubMed dictionaries back to articles class.
    for i, pub_list in enumerate(query_json["PubMed"]):
        new_list = []
        for pub in pub_list:
            new_list.append(pymed.article.PubMedArticle(ET.fromstring(pub["xml"])))
        query_json["PubMed"][i] = new_list
    return query_json


def test_input_reading_and_checking_no_prev_pub(config_dict):
    config_json_filepath = os.path.join("tests", "testing_files", "config_truncated.json")
    ref_path_or_URL = os.path.join("tests", "testing_files", "tokenized_citations_duplicates_removed.json")
    
    expected_config_dict = config_dict
    
    actual_config_dict, tokenized_citations, has_previous_pubs, prev_pubs = \
        input_reading_and_checking(config_json_filepath, ref_path_or_URL, False, False, False, "ignore", True)
        
    assert expected_config_dict == actual_config_dict
    assert tokenized_citations == load_json(os.path.join("tests", "testing_files", "tokenized_citations_duplicates_removed.json"))
    assert has_previous_pubs == False
    assert prev_pubs == {}



def test_input_reading_and_checking_has_prev_pub(config_dict):
    config_json_filepath = os.path.join("tests", "testing_files", "config_truncated.json")
    ref_path_or_URL = os.path.join("tests", "testing_files", "tokenized_citations_duplicates_removed.json")
    prev_pub_filepath = os.path.join("tests", "testing_files", "publication_dict_truncated.json")
    
    expected_config_dict = config_dict
    
    actual_config_dict, tokenized_citations, has_previous_pubs, prev_pubs = \
        input_reading_and_checking(config_json_filepath, ref_path_or_URL, False, False, False, prev_pub_filepath, True)
    
    assert expected_config_dict == actual_config_dict
    assert tokenized_citations == load_json(os.path.join("tests", "testing_files", "tokenized_citations_duplicates_removed.json"))
    assert has_previous_pubs == True
    assert prev_pubs == load_json(os.path.join("tests", "testing_files", "publication_dict_truncated.json"))
    
    
def test_input_reading_and_checking_noCrossref(config_dict):
    config_json_filepath = os.path.join("tests", "testing_files", "config_truncated_noCrossref.json")
    ref_path_or_URL = os.path.join("tests", "testing_files", "tokenized_citations_duplicates_removed.json")
    
    expected_config_dict = load_json(os.path.join("tests", "testing_files", "config_truncated_noCrossref.json"))
    
    actual_config_dict, tokenized_citations, has_previous_pubs, prev_pubs = \
        input_reading_and_checking(config_json_filepath, ref_path_or_URL, False, False, False, "ignore", True)
    
    assert expected_config_dict == actual_config_dict
    assert tokenized_citations == load_json(os.path.join("tests", "testing_files", "tokenized_citations_duplicates_removed.json"))
    assert has_previous_pubs == False
    assert prev_pubs == {}
    

def test_input_reading_and_checking_noPubMed(config_dict):
    config_json_filepath = os.path.join("tests", "testing_files", "config_truncated_noPubMed.json")
    ref_path_or_URL = os.path.join("tests", "testing_files", "tokenized_citations_duplicates_removed.json")
    
    expected_config_dict = load_json(os.path.join("tests", "testing_files", "config_truncated_noPubMed.json"))
    
    actual_config_dict, tokenized_citations, has_previous_pubs, prev_pubs = \
        input_reading_and_checking(config_json_filepath, ref_path_or_URL, False, False, False, "ignore", True)
    
    assert expected_config_dict == actual_config_dict
    assert tokenized_citations == load_json(os.path.join("tests", "testing_files", "tokenized_citations_duplicates_removed.json"))
    assert has_previous_pubs == False
    assert prev_pubs == {}




def test_build_publication_dict_with_Crossref(mocker, config_dict_Hunter_only, original_queries, tokenized_citations):
    
    # running_pubs = {}
    # running_pubs1, matching_key_for_citation1, all_pubs = search_references_on_source("PubMed", running_pubs, tokenized_citations, "asdf", original_queries["PubMed"])
    # running_pubs2, matching_key_for_citation2, all_pubs = search_references_on_source("Crossref", copy.deepcopy(running_pubs1), tokenized_citations, "asdf", original_queries["Crossref"])
    
    # running_pubs3, matching_key_for_citation3, all_pubs = search_references_on_source("PubMed", copy.deepcopy(running_pubs2), tokenized_citations, "asdf", original_queries["PubMed"])
    # running_pubs4, matching_key_for_citation4, all_pubs = search_references_on_source("Crossref", copy.deepcopy(running_pubs3), tokenized_citations, "asdf", original_queries["Crossref"])

    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "running_pubs1.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "running_pubs2.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "running_pubs3.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs3, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "running_pubs4.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs4, indent=2, sort_keys=True))
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "matching_key_for_citation1.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(matching_key_for_citation1, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "matching_key_for_citation2.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(matching_key_for_citation2, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "matching_key_for_citation3.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(matching_key_for_citation3, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "matching_key_for_citation4.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(matching_key_for_citation4, indent=2, sort_keys=True))
    
    running_pubs1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "running_pubs1.json"))
    running_pubs2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "running_pubs2.json"))
    running_pubs3 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "running_pubs3.json"))
    running_pubs4 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "running_pubs4.json"))
    
    matching_key_for_citation1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "matching_key_for_citation1.json"))
    matching_key_for_citation2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "matching_key_for_citation2.json"))
    matching_key_for_citation3 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "matching_key_for_citation3.json"))
    matching_key_for_citation4 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "matching_key_for_citation4.json"))
    
    mocker.patch("academic_tracker.ref_srch_modularized.ref_srch_webio.search_references_on_source", 
                  side_effect=[(running_pubs1, matching_key_for_citation1, original_queries["PubMed"]), 
                                (running_pubs2, matching_key_for_citation2, original_queries["Crossref"]),
                                (running_pubs3, matching_key_for_citation3, original_queries["PubMed"]),
                                (running_pubs4, matching_key_for_citation4, original_queries["Crossref"])])    
    
    actual_publication_dict, actual_tokenized_citations, _ = build_publication_dict(config_dict_Hunter_only, tokenized_citations, False, False)
    
    ## Have to sort to match how it is saved as JSON.
    actual_publication_dict = {key:{key2:value2 for key2, value2 in sorted(value.items())} for key, value in sorted(actual_publication_dict.items())}
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "test_tok_cits.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_tokenized_citations, indent=2, sort_keys=True))
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "test_pub_dict.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_publication_dict, indent=2, sort_keys=True))
    
    expected_pub_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "publication_dict.json"))
    expected_tokenized_citations = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "all", "tokenized_reference.json"))
    
    assert expected_pub_dict == actual_publication_dict
    assert expected_tokenized_citations == actual_tokenized_citations
    


def test_build_publication_dict_no_Crossref(mocker, config_dict_Hunter_only, original_queries, tokenized_citations):
    
    # running_pubs = {}
    # running_pubs1, matching_key_for_citation1, all_pubs = search_references_on_source("PubMed", running_pubs, tokenized_citations, "asdf", original_queries["PubMed"])
    
    # running_pubs2, matching_key_for_citation2, all_pubs = search_references_on_source("PubMed", copy.deepcopy(running_pubs1), tokenized_citations, "asdf", original_queries["PubMed"])

    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "running_pubs1.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "running_pubs2.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "matching_key_for_citation1.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(matching_key_for_citation1, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "matching_key_for_citation2.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(matching_key_for_citation2, indent=2, sort_keys=True))
    
    running_pubs1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "running_pubs1.json"))
    running_pubs2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "running_pubs2.json"))
    
    matching_key_for_citation1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "matching_key_for_citation1.json"))
    matching_key_for_citation2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "matching_key_for_citation2.json"))
    
    mocker.patch("academic_tracker.ref_srch_modularized.ref_srch_webio.search_references_on_source", 
                  side_effect=[(running_pubs1, matching_key_for_citation1, original_queries["PubMed"]), 
                                (running_pubs2, matching_key_for_citation2, original_queries["PubMed"])])
        
    
    actual_publication_dict, actual_tokenized_citations, _ = build_publication_dict(config_dict_Hunter_only, tokenized_citations, True, False)
    
    ## Have to sort to match how it is saved as JSON.
    actual_publication_dict = {key:{key2:value2 for key2, value2 in sorted(value.items())} for key, value in sorted(actual_publication_dict.items())}
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "test_tok_cits.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_tokenized_citations, indent=2, sort_keys=True))
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "test_pub_dict.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_publication_dict, indent=2, sort_keys=True))
    
    expected_pub_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "publication_dict.json"))
    expected_tokenized_citations = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_Crossref", "tokenized_reference.json"))
    
    assert expected_pub_dict == actual_publication_dict
    assert expected_tokenized_citations == actual_tokenized_citations
    
    


def test_build_publication_dict_no_PubMed(mocker, config_dict_Hunter_only, original_queries, tokenized_citations):
    
    # running_pubs = {}
    # running_pubs1, matching_key_for_citation1, all_pubs = search_references_on_source("Crossref", running_pubs, tokenized_citations, "asdf", original_queries["Crossref"])
    
    # running_pubs2, matching_key_for_citation2, all_pubs = search_references_on_source("Crossref", copy.deepcopy(running_pubs1), tokenized_citations, "asdf", original_queries["Crossref"])

    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "running_pubs1.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs1, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "running_pubs2.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(running_pubs2, indent=2, sort_keys=True))
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "matching_key_for_citation1.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(matching_key_for_citation1, indent=2, sort_keys=True))
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "matching_key_for_citation2.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(matching_key_for_citation2, indent=2, sort_keys=True))
    
    running_pubs1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "running_pubs1.json"))
    running_pubs2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "running_pubs2.json"))
    
    matching_key_for_citation1 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "matching_key_for_citation1.json"))
    matching_key_for_citation2 = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "matching_key_for_citation2.json"))
        
    mocker.patch("academic_tracker.ref_srch_modularized.ref_srch_webio.search_references_on_source", 
                  side_effect=[(running_pubs1, matching_key_for_citation1, original_queries["Crossref"]), 
                                (running_pubs2, matching_key_for_citation2, original_queries["Crossref"])])
    
    
    actual_publication_dict, actual_tokenized_citations, _ = build_publication_dict(config_dict_Hunter_only, tokenized_citations, False, True)
    
    ## Have to sort to match how it is saved as JSON.
    actual_publication_dict = {key:{key2:value2 for key2, value2 in sorted(value.items())} for key, value in sorted(actual_publication_dict.items())}
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "test_tok_cits.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_tokenized_citations, indent=2, sort_keys=True))
    
    # with open(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "test_pub_dict.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_publication_dict, indent=2, sort_keys=True))
    
    expected_pub_dict = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "publication_dict.json"))
    expected_tokenized_citations = load_json(os.path.join("tests", "testing_files", "intermediate_results", "ref_search", "no_PubMed", "tokenized_reference.json"))
    
    assert expected_pub_dict == actual_publication_dict
    assert expected_tokenized_citations == actual_tokenized_citations




def test_save_and_send_reports_and_emails_no_email(config_dict):
    
    tokenized_citations = load_json(os.path.join("tests", "testing_files", "tokenized_citations_for_report_test.json"))
    pub_dict = load_json(os.path.join("tests", "testing_files", "ref_srch_Crossref_pub_dict.json"))
    
    config_dict["summary_report"] = {}
    config_dict["summary_report"]["template"] = read_text_from_txt(os.path.join("tests", "testing_files", "ref_srch_report_template_string.txt"))
    
    save_dir = save_and_send_reports_and_emails(config_dict, tokenized_citations, pub_dict, {}, False, False)
    
    assert os.path.exists(os.path.join(save_dir, "summary_report.txt"))
    assert read_text_from_txt(os.path.join("tests", "testing_files", "ref_srch_report_test1.txt")) == read_text_from_txt(os.path.join(save_dir, "summary_report.txt"))
    assert not os.path.exists(os.path.join(save_dir, "emails.json"))



def test_save_and_send_reports_and_emails_with_email(config_dict):
    
    tokenized_citations = load_json(os.path.join("tests", "testing_files", "tokenized_citations_for_report_test.json"))
    pub_dict = load_json(os.path.join("tests", "testing_files", "ref_srch_Crossref_pub_dict.json"))
    
    config_dict["summary_report"] = {}
    config_dict["summary_report"]["template"] = read_text_from_txt(os.path.join("tests", "testing_files", "ref_srch_report_template_string.txt"))
    config_dict["summary_report"]["from_email"] = "ptth222@uky.edu"
    config_dict["summary_report"]["to_email"] = ["ptth222@uky.edu"]
    config_dict["summary_report"]["cc_email"] = ["ptth222@uky.edu"]
    config_dict["summary_report"]["email_body"] = "Body"
    config_dict["summary_report"]["email_subject"] = "Subject"
    
    save_dir = save_and_send_reports_and_emails(config_dict, tokenized_citations, pub_dict, {}, False, True)
    
    assert os.path.exists(os.path.join(save_dir, "summary_report.txt"))
    assert read_text_from_txt(os.path.join("tests", "testing_files", "ref_srch_report_test1.txt")) == read_text_from_txt(os.path.join(save_dir, "summary_report.txt"))
    assert os.path.exists(os.path.join(save_dir, "emails.json"))



def test_save_and_send_reports_and_emails_default_template(config_dict):
    
    tokenized_citations = load_json(os.path.join("tests", "testing_files", "tokenized_citations_for_report_test.json"))
    pub_dict = load_json(os.path.join("tests", "testing_files", "ref_srch_Crossref_pub_dict.json"))
    
    config_dict["summary_report"] = {}
    
    save_dir = save_and_send_reports_and_emails(config_dict, tokenized_citations, pub_dict, {}, False, False)
    # with open(os.path.join("tests", "testing_files", "ref_srch_report_default_new.txt"), 'wb') as outFile:
    #     outFile.write(read_text_from_txt(os.path.join(save_dir, "summary_report.txt")).encode("utf-8"))
    
    assert os.path.exists(os.path.join(save_dir, "summary_report.txt"))
    assert read_text_from_txt(os.path.join("tests", "testing_files", "ref_srch_report_default.txt")) == read_text_from_txt(os.path.join(save_dir, "summary_report.txt"))
    assert not os.path.exists(os.path.join(save_dir, "emails.json"))
    
    
def test_save_and_send_reports_and_emails_tabular(config_dict):
    
    tokenized_citations = load_json(os.path.join("tests", "testing_files", "tokenized_citations_for_report_test.json"))
    pub_dict = load_json(os.path.join("tests", "testing_files", "ref_srch_Crossref_pub_dict.json"))
    
    config_dict["summary_report"] = {}
    config_dict["summary_report"]["columns"] = {"Col1":"<author_first>"}
    
    save_dir = save_and_send_reports_and_emails(config_dict, tokenized_citations, pub_dict, {}, False, False)
    
    assert os.path.exists(os.path.join(save_dir, "summary_report.csv"))
    assert not os.path.exists(os.path.join(save_dir, "emails.json"))
    
    
def test_save_and_send_reports_and_emails_manual_name(config_dict):
    
    tokenized_citations = load_json(os.path.join("tests", "testing_files", "tokenized_citations_for_report_test.json"))
    pub_dict = load_json(os.path.join("tests", "testing_files", "ref_srch_Crossref_pub_dict.json"))
    
    config_dict["summary_report"] = {}
    config_dict["summary_report"]["template"] = read_text_from_txt(os.path.join("tests", "testing_files", "ref_srch_report_template_string.txt"))
    config_dict["summary_report"]["filename"] = "test_name.txt"
    
    save_dir = save_and_send_reports_and_emails(config_dict, tokenized_citations, pub_dict, {}, False, False)
    
    assert os.path.exists(os.path.join(save_dir, "test_name.txt"))
    assert not os.path.exists(os.path.join(save_dir, "emails.json"))









@pytest.fixture(autouse=True)
def cleanup(request):
    """Cleanup a testing directory once we are finished."""
    def remove_test_dir():
        dir_contents = os.listdir()
        tracker_dirs = [folder for folder in dir_contents if re.match(r"tracker-(\d{10})", folder) or re.match(r"tracker-test-(\d{10})", folder)]
        for directory in tracker_dirs:
            shutil.rmtree(directory)
    request.addfinalizer(remove_test_dir)








# -*- coding: utf-8 -*-


import os
import time
import json

import pytest
import shutil

from academic_tracker.emails_and_reports_helpers import _replace_keywords
from academic_tracker.fileio import load_json


TESTING_DIR = "test_dir"
@pytest.fixture(scope="module", autouse=True)
def test_email_dir():
    save_dir_name = TESTING_DIR
    os.mkdir(save_dir_name)
    time_to_wait = 10
    time_counter = 0
    
    while not os.path.exists(save_dir_name):
        time.sleep(1)
        time_counter += 1
        if time_counter > time_to_wait:
            raise FileNotFoundError(save_dir_name + " was not created within " + str(time_to_wait) + " seconds, so it is assumed that it won't be and something went wrong.")





@pytest.fixture
def tokenized_citations():
    return load_json(os.path.join("tests", "testing_files", "tokenized_citations_for_report_test.json"))


@pytest.fixture
def publication_dict():
    return load_json(os.path.join("tests", "testing_files", "ref_srch_Crossref_pub_dict.json"))


@pytest.fixture
def config_dict():
    config_dict = {"summary_report":{"columns":{'Authors': '<authors>',
                                                'Grants': '<grants>',
                                                'Abstract': '<abstract>',
                                                'Conclusions': '<conclusions>',
                                                'Copyrights': '<copyrights>',
                                                'DOI': '<DOI>',
                                                'Journal': '<journal>',
                                                'Keywords': '<keywords>',
                                                'Methods': '<methods>',
                                                'PMID': '<PMID>',
                                                'Results': '<results>',
                                                'Title': '<title>',
                                                'PMCID': '<PMCID>',
                                                'Publication Year': '<publication_year>',
                                                'Publication Month': '<publication_month>',
                                                'Publication Day': '<publication_day>',
                                                'Tok Title': '<tok_title>',
                                                'Tok DOI': '<tok_DOI>',
                                                'Tok PMID': '<tok_PMID>',
                                                'Tok Authors': '<tok_authors>',
                                                'Ref Line': '<ref_line>',
                                                'Comparison': '<is_in_comparison_file>',
                                                'First Author': '<first_author>',
                                                'Last Author': '<last_author>',
                                                'Pub_Authors': '<pub_author_last>, <pub_author_first> <pub_author_initials> <pub_author_affiliations>'},
                                     "column_order":['Authors', 'Grants', 'Abstract', 'Conclusions', 'Copyrights', 'DOI', 'Journal', 'Keywords', 'Methods',
                                                     'PMID', 'Results', 'Title', 'PMCID', 'Publication Year', 'Publication Month', 'Publication Day',
                                                     'Tok Title', 'Tok DOI', 'Tok PMID', 'Tok Authors', 'Ref Line', 'Comparison', 'First Author', 
                                                     'Last Author', 'Pub_Authors'],
                                     "sort":["Authors"],
                                     "file_format":"csv",
                                     "filename":"test_name.csv",
                                     "separator":"\t"}}
    
    return config_dict


def test_replace_keywords1(publication_dict, config_dict, tokenized_citations):
    
    template = config_dict["summary_report"]["columns"]
    
    expected_template = {'Authors': 'Huan Jin, Joshua M. Mitchell, Hunter N. B. Moseley',
                          'Grants': '1419282',
                          'Abstract': 'None',
                          'Conclusions': 'None',
                          'Copyrights': 'None',
                          'DOI': '10.3390/metabo10090368',
                          'Journal': 'MDPI AG',
                          'Keywords': 'None',
                          'Methods': 'None',
                          'PMID': 'None',
                          'Results': 'None',
                          'Title': 'Atom Identifiers Generated by a Neighborhood-Specific Graph Coloring Method Enable Compound Harmonization across Metabolic Databases',
                          'PMCID': 'None',
                          'Publication Year': '2020',
                          'Publication Month': '9',
                          'Publication Day': '11',
                          'Tok Title': 'Atom Identifiers Generated by a Neighborhood-Specific Graph Coloring Method Enable Compound Harmonization across Metabolic Databases.',
                          'Tok DOI': '10.3390/metabo10090368',
                          'Tok PMID': 'None',
                          'Tok Authors': 'Jin H, Mitchell J, Moseley H',
                          'Ref Line': 'Jin H, Mitchell J, Moseley H.  Atom Identifiers Generated by a Neighborhood-Specific Graph Coloring Method Enable Compound Harmonization across Metabolic Databases. Metabolites. 2020 September; 10(9):368-. doi: 10.3390/metabo10090368.',
                          'Comparison': 'N/A',
                          'First Author': 'Jin, Huan',
                          'Last Author': 'Moseley, Hunter N. B.',
                          'Pub_Authors': '<pub_author_last>, <pub_author_first> <pub_author_initials> <pub_author_affiliations>'}
    
    actual_template = _replace_keywords(template, 
                                        publication_dict, 
                                        None, 
                                        pub='https://doi.org/10.3390/metabo10090368', 
                                        tokenized_citation=tokenized_citations[1], 
                                        is_citation_in_prev_pubs=None, 
                                        pub_author={})
    # with open(os.path.join("tests", "testing_files", "aaa.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_template, indent=2, sort_keys=False))
    
    assert expected_template == actual_template
    

    
def test_replace_keywords2(publication_dict, config_dict, tokenized_citations):
    
    template = config_dict["summary_report"]["columns"]
    publication_dict['https://doi.org/10.3390/metabo10090368']["grants"] = ["asdf"]
    tokenized_citations[1]["reference_line"] = ""
    
    expected_template = {'Authors': 'Huan Jin, Joshua M. Mitchell, Hunter N. B. Moseley',
                          'Grants': 'asdf',
                          'Abstract': 'None',
                          'Conclusions': 'None',
                          'Copyrights': 'None',
                          'DOI': '10.3390/metabo10090368',
                          'Journal': 'MDPI AG',
                          'Keywords': 'None',
                          'Methods': 'None',
                          'PMID': 'None',
                          'Results': 'None',
                          'Title': 'Atom Identifiers Generated by a Neighborhood-Specific Graph Coloring Method Enable Compound Harmonization across Metabolic Databases',
                          'PMCID': 'None',
                          'Publication Year': '2020',
                          'Publication Month': '9',
                          'Publication Day': '11',
                          'Tok Title': 'Atom Identifiers Generated by a Neighborhood-Specific Graph Coloring Method Enable Compound Harmonization across Metabolic Databases.',
                          'Tok DOI': '10.3390/metabo10090368',
                          'Tok PMID': 'None',
                          'Tok Authors': 'Jin H, Mitchell J, Moseley H',
                          'Ref Line': 'N/A',
                          'Comparison': 'True',
                          'First Author': 'Jin, Huan',
                          'Last Author': 'Moseley, Hunter N. B.',
                          'Pub_Authors': 'Jin, Huan None None'}
    
    actual_template = _replace_keywords(template, 
                                        publication_dict,
                                        None,
                                        pub='https://doi.org/10.3390/metabo10090368', 
                                        tokenized_citation=tokenized_citations[1], 
                                        is_citation_in_prev_pubs=True, 
                                        pub_author=publication_dict['https://doi.org/10.3390/metabo10090368']["authors"][0])
    # with open(os.path.join("tests", "testing_files", "aaa.json"),'w') as jsonFile:
    #     jsonFile.write(json.dumps(actual_template, indent=2, sort_keys=False))
    
    assert expected_template == actual_template





@pytest.fixture
def publication_dict2():
    return load_json(os.path.join("tests", "testing_files", "publication_dict_truncated.json"))

@pytest.fixture
def config_dict2():
    return load_json(os.path.join("tests", "testing_files", "config_truncated.json"))


def test_replace_keywords(publication_dict2, config_dict2):
    
    template = {"Col1":"<project_name>", "Col2":"<author_first>", "Col3":"<title>", "Col4":"<pub_author_first>"}
    
    expected_template = {"Col1":"Project 1", "Col2":"Hunter", "Col3":"Identifying and sharing per-and polyfluoroalkyl substances hot-spot areas and exposures in drinking water.", "Col4":"Travis"}
    
    pub_author = {"firstname":"Travis",
                  "lastname":"Thompson",
                  "initials":"P.T.T.",
                  "affiliation":"University of Kentucky",
                  "ORCID": "0000-0002-8198-1327",
                  "author_id": "Travis Thompson"}
    
    actual_template = _replace_keywords(template, 
                                        publication_dict2, 
                                        config_dict2, 
                                        project_name="Project 1", 
                                        author="Hunter Moseley", 
                                        pub="https://doi.org/10.1038/s41597-023-02277-x", 
                                        pub_author=pub_author)
    
    assert expected_template == actual_template
    

def test_replace_keywords_more_keywords(publication_dict2, config_dict2):
    
    template = {"Col1":"<first_author>", "Col2":"<last_author>", "Col3":"<authors>", "Col4":"<grants>", "Col5":"<publication_year>"}
    
    expected_template = {"Col1":'Ojha, Sweta', "Col2":'Pennell, Kelly G', "Col3":'Sweta Ojha, P Travis Thompson, Christian D Powell, Hunter N B Moseley, Kelly G Pennell', "Col4":'P42 ES007380, 2020026', "Col5":"2023"}
    
    pub_author = {"firstname":"Travis",
                  "lastname":"Thompson",
                  "initials":"P.T.T.",
                  "affiliation":"University of Kentucky",
                  "ORCID": "0000-0002-8198-1327",
                  "author_id": "Travis Thompson"}
    
    actual_template = _replace_keywords(template, 
                                        publication_dict2, 
                                        config_dict2, 
                                        project_name="Project 1", 
                                        author="Hunter Moseley", 
                                        pub="https://doi.org/10.1038/s41597-023-02277-x", 
                                        pub_author=pub_author)
    
    assert expected_template == actual_template
    
    
    
def test_replace_keywords_no_change(publication_dict2, config_dict2):
    
    template = {"Col1":"<project_name>", "Col2":"<author_first>", "Col3":"<title>", "Col4":"<pub_author_first>"}
    
    expected_template = template
    
    actual_template = _replace_keywords(template, 
                                        publication_dict2, 
                                        config_dict2)
    
    assert expected_template == actual_template







@pytest.fixture(scope="module", autouse=True)
def cleanup(request):
    """Cleanup a testing directory once we are finished."""
    def remove_test_dir():
        if os.path.exists(TESTING_DIR):
            shutil.rmtree(TESTING_DIR)
    request.addfinalizer(remove_test_dir)




# -*- coding: utf-8 -*-
"""
Helper Functions
~~~~~~~~~~~~~~~~

This module contains helper functions, such as printing, and regex searching.
"""

import re
import copy
import collections.abc
import xml.etree.ElementTree as ET

import fuzzywuzzy.fuzz

from . import __main__
from . import webio

DOI_URL = webio.DOI_URL
PUBLICATION_TEMPLATE = webio.PUBLICATION_TEMPLATE


def vprint(*args, verbosity=0):
    """Print depending on the state of VERBOSE, SILENT, and verbosity.
    
    If the global SILENT is True don't print anything. If verbosity is 0 then 
    print. If verbosity is 1 then VERBOSE must be True to print.
    
    Args:
        verbosity (int): Either 0 or 1 for different levels of verbosity.
    """
    
    VERBOSE = __main__.VERBOSE
    SILENT = __main__.SILENT
    
    if SILENT:
        return
    
    if verbosity == 0:
        print(*args)
    elif VERBOSE and verbosity == 1:
        print(*args)



def create_authors_by_project_dict(config_dict):
    """Create the authors_by_project_dict dict from the config_dict.
    
    Creates a dict where the keys are the projects in the config_dict and the values
    are the authors associated with that project from config_dict["Authors"].
    
    Args:
        config_dict (dict): schema matches the JSON Project Tracking Configuration file.
        
    Returns:
        authors_by_project_dict (dict): keys are the projects in the config_dict and the values are the authors associated with that project from config_dict["Authors"].
    """
    
    authors_by_project_dict = {project:{} for project in config_dict["project_descriptions"]}
    for project, project_attributes in config_dict["project_descriptions"].items():
        if "authors" in project_attributes:
            for author in project_attributes["authors"]:
                if author in config_dict["Authors"]:
                    authors_by_project_dict[project][author] = copy.deepcopy(config_dict["Authors"][author])
                else:
                    vprint("Warning: The author, " + author + ", in the " + project + " project of the project tracking configuration file could not be found in the Authors section of the Configuration JSON file.")
        else:
            authors_by_project_dict[project] = copy.deepcopy(config_dict["Authors"])
    
        for author_attr in authors_by_project_dict[project].values():
            for key in project_attributes:
                author_attr.setdefault(key, project_attributes[key])
                
    return authors_by_project_dict


def adjust_author_attributes(authors_by_project_dict, config_dict):
    """Modifies config_dict with values from authors_by_project_dict
    
    Go through the authors in authors_by_project_dict and find the lowest cutoff_year.
    Also find affiliations and grants and create a union of them across projects.
    Update the authors in config_dict["Authors"].
    
    Args:
        authors_by_project_dict (dict): keys are the projects in the config_dict and the values are the authors associated with that project from config_dict["Authors"].
        config_dict (dict): schema matches the JSON Project Tracking Configuration file.
        
    Returns:
        config_dict (dict): schema matches the JSON Project Tracking Configuration file.
    """
    
    for author, author_attr in config_dict["Authors"].items():
        cutoff_year = author_attr["cutoff_year"] if "cutoff_year" in author_attr else 9999
        
        affiliations = set(author_attr["affiliations"]) if "affiliations" in author_attr else set()
        
        grants = set(author_attr["grants"]) if "grants" in author_attr else set()
        
        for project, authors_dict in authors_by_project_dict.items():
            if author in authors_dict:
                if "cutoff_year" in authors_dict[author] and authors_dict[author]["cutoff_year"] < cutoff_year:
                    cutoff_year = authors_dict[author]["cutoff_year"]
                    
                if "affiliations" in authors_dict[author]:
                    affiliations |= set(authors_dict[author]["affiliations"])
                    
                if "grants" in authors_dict[author]:
                    grants |= set(authors_dict[author]["grants"])
                    
                if "collaborator_report" in authors_dict[author]:
                    config_dict["Authors"][author]["collaborator_report"] = authors_dict[author]["collaborator_report"]
                    
        affiliations = list(affiliations)
        grants = list(grants)
        grants.sort()
        affiliations.sort()
        config_dict["Authors"][author]["cutoff_year"] = cutoff_year
        config_dict["Authors"][author]["affiliations"] = affiliations
        config_dict["Authors"][author]["grants"] = grants
        
    return config_dict



    
def regex_match_return(regex, string_to_match):
    """Return the groups matched in the regex if the regex matches.
    
    regex is delivered to re.match() with string_to_match, and if there is a match 
    the match.groups() is returned, otherwise an empty tuple is returned.
    
    Args:
        regex (str): A string with a regular expression to be delivered to re.match().
        string_to_match (str): The string to match with the regex.
        
    Returns:
        (tuple): either the tuple of the matched groups in the regex or an empty tuple if a match wasn't found.
    """
    
    match = re.match(regex, string_to_match)
    return match.groups() if match else ()



def regex_group_return(regex_groups, group_index):
    """Return the group in the regex_groups indicated by group_index if it exists, else return empty string.
    
    If group_index is out of range of the regex_groups an empty string is retruned.
    
    Args:
        regex_groups (tuple): A tuple returned from a matched regex.groups() call.
        group_number (int): The index of the regex_groups to return.
        
    Returns:
        (str): Either emtpy string or the group string matched by the regex.
    """
    
    return regex_groups[group_index] if group_index < len(regex_groups) and regex_groups[group_index] else ""



def regex_search_return(regex, string_to_search):
    """Return the groups matched in the regex if the regex matches.
    
    regex is delivered to re.search() with string_to_search, and if there is a match 
    the match.groups() is returned, otherwise an empty tuple is returned.
    
    Args:
        regex (str): A string with a regular expression to be delivered to re.search().
        string_to_search (str): The string to match with the regex.
    
    Returns:
        (tuple): either the tuple of the matched groups in the regex or an empty tuple if a match wasn't found.
    """
    
    match = re.search(regex, string_to_search)
    return match.groups() if match else ()


def _generate_first_name_match_regex(name):
    """Generate a regular expression to match the given first name.
    
    Args:
        name (str): first name to generate regex for.
    
    Returns:
        (str) regular expression to use with re.match().
    """
    return ".* " + name + "|" + name + " .*|" + name


def do_strings_fuzzy_match(string1, string2, match_ratio=90):
    """Fuzzy match the 2 strings and if the ratio is greater than or equal to match_ratio, return True.
    
    Args:
        string1 (str|None): a string to fuzzy match.
        string2 (str|None): a string to fuzzy match.
        match_ratio (int): the ratio (0-100) that the match must be greater than to return True.
    
    Returns:
        (bool): True if strings match, False otherwise.
    """
    if string1 is None or string2 is None:
        return False
    
    string1 = string1.lower()
    string2 = string2.lower()
    if fuzzywuzzy.fuzz.ratio(string1, string2) >= match_ratio or\
       fuzzywuzzy.fuzz.ratio(string2, string1) >= match_ratio:
           return True
    return False



    
# def match_authors_in_pub_PubMed(authors_json, author_list):
#     """Look for matching authors in PubMed pub data.
    
#     Goes through the author list from PubMed and tries matching to an author in 
#     authors_json using firstname, lastname, and affiliations.
    
#     Args:
#         authors_json (dict): keys are authors and values are author attributes. Matches authors JSON schema.
#         author_list (list): list of dicts where each dict is attributes of an author.
        
#     Returns:
#         author_list (list): either the author list with matched authors containing an additional author_id attribute, or an empty list if no authors were matched.
#     """
    
#     ## pub.authors are dictionaries with lastname, firstname, initials, and affiliation. firstname can have initials at the end ex "Andrew P"
#     publication_has_affiliated_author = False
#     for author_items in author_list:
#         author_items_affiliation = str(author_items.get("affiliation")).lower()
#         author_items_first_name = str(author_items.get("firstname")).replace(".","").lower()
#         author_items_last_name = str(author_items.get("lastname")).replace(".","").lower()
        
#         for author, author_attributes in authors_json.items():
#             author_json_first_name = author_attributes["first_name"].replace(".","").lower()
#             author_json_last_name = author_attributes["last_name"].replace(".","").lower()
        
#             ## Match the author's first and last name and then match affiliations.
#             ## The first name is matched with an additional .* to try and allow for the addition of initials, but this could cause bad matches. 
#             ## For example the name Hu will match Hubert. Counting on the last name to reduce errors.
#             ## Note that the PubMed query will return publications where the author is just a collaborater, so not finding a match in the authors isn't uncommon.
#             if re.match(_generate_first_name_match_regex(author_json_first_name), author_items_first_name) and \
#                author_json_last_name == author_items_last_name:
#                 ## affiliations are sets of strings so the intersection should have at least 1 string for a match.
                
#                 if any([affiliation.lower() in author_items_affiliation for affiliation in author_attributes["affiliations"]]):
#                     publication_has_affiliated_author = True
#                     author_items["author_id"] = author
#                     break
                
#     return author_list if publication_has_affiliated_author else []


    
    
# def match_authors_in_pub_Crossref(authors_json, author_list):
#     """Look for matching authors in Crossref pub data.
    
#     Goes through the author list from Crossref and tries matching to an author in 
#     authors_json using firstname, lastname, and affiliations, or ORCID if ORCID
#     is present.
    
#     Args:
#         authors_json (dict): keys are authors and values are author attributes. Matches authors JSON schema.
#         author_list (list): list of dicts where each dict is attributes of an author.
        
#     Returns:
#         author_list (list): either the author list with matched authors containing an additional author_id attribute, or an empty list if no authors were matched.
#     """
    
#     publication_has_affiliated_author = False
#     for author_items in author_list:
#         ## If the author has no affiliation or ORCID id then go to the next.
#         if not author_items["affiliation"] and not "ORCID" in author_items:
#             continue
        
#         can_name_match = True
#         if author_items.get("given") is None:
#             can_name_match = False
#         else:
#             author_items_first_name = author_items["given"].replace(".","").lower()
        
#         if author_items.get("family") is None:
#             can_name_match = False
#         else:
#             author_items_last_name = author_items["family"].replace(".","").lower()
        
#         if author_items["affiliation"] and "name" in author_items["affiliation"][0]:
#             author_items_affiliation = author_items["affiliation"][0]["name"].lower()
#         else:
#             author_items_affiliation = []
                
#         if "ORCID" in author_items:
#             ORCID_id = regex_search_return(r"(\d{4}-\d{4}-\d{4}-\d{3}[0,1,2,3,4,5,6,7,8,9,X])", author_items["ORCID"])[0]
#         else:
#             ORCID_id = ""
        
#         for author, author_attributes in authors_json.items():
#             author_json_first_name = author_attributes["first_name"].replace(".","").lower()
#             author_json_last_name = author_attributes["last_name"].replace(".","").lower()
        
#             ## Match the author's first and last name and then match affiliations.
#             ## The first name is matched with an additional .* to try and allow for the addition of initials, but this could cause bad matches. 
#             ## For example the name Hu will match Hubert. Counting on the last name to reduce errors.
#             ## Note that the PubMed query will return publications where the author is just a collaborater, so not finding a match in the authors isn't uncommon.
#             if can_name_match and re.match(_generate_first_name_match_regex(author_json_first_name), author_items_first_name) and \
#                author_json_last_name == author_items_last_name:
#                 ## affiliations are sets of strings so the intersection should have at least 1 string for a match.
#                 if any([affiliation.lower() in author_items_affiliation for affiliation in author_attributes["affiliations"]]):
#                     publication_has_affiliated_author = True
#                     author_items["author_id"] = author
#                     break
                
#             elif ORCID_id and "ORCID" in author_attributes and ORCID_id == author_attributes["ORCID"]:
#                 publication_has_affiliated_author = True
#                 author_items["author_id"] = author
#                 break
                
#     if publication_has_affiliated_author:
#         return author_list
#     else:
#         return []



def match_pub_authors_to_citation_authors(citation_authors, author_list):
    """Try to match authors in pub data to authors in citation data.
    
    Goes through the author_list from a publication and tries matching to an author in 
    citation_authors using last name or ORCID if ORCID is present.
    
    Args:
        citation_authors (list): list of dicts where each dict can have last name and ORCID or collective_name and ORCID.
        author_list (list): list of dicts where each dict is the attributes of an author.
        
    Returns:
        (bool): True if an author was matched, False otherwise.
    """
    for author_items in author_list:
        is_collective_author = False
        has_collective_name = False
        if "collectivename" in author_items:
            is_collective_author = True
            if author_items["collectivename"] is not None:
                has_collective_name = True
                author_items_collective_name = author_items["collectivename"]
        
        else:
            can_name_match = True
            if author_items["lastname"] is None:
                can_name_match = False
            else:
                author_items_last_name = author_items["lastname"].replace(".","").lower()
                                    
        for author_attributes in citation_authors:
            ## Try to match on ORCID.
            if author_items["ORCID"] and "ORCID" in author_attributes and author_items["ORCID"] == author_attributes["ORCID"]:
                return True
            
            citation_is_collective_author = False
            citation_has_collective_name = False
            if "collective_name" in author_attributes:
                citation_is_collective_author = True
                if author_attributes["collective_name"] is not None:
                    citation_has_collective_name = True
                    citation_author_collective_name = author_attributes["collective_name"]
            
            ## If it is a collective author then match on collectivename, else match on first and last.
            if is_collective_author:
                if not has_collective_name or \
                   not citation_is_collective_author or \
                   not citation_has_collective_name:
                    continue
                
                if do_strings_fuzzy_match(citation_author_collective_name, author_items_collective_name):
                       return True
            
            elif not citation_is_collective_author:
                citation_author_last_name = author_attributes["last"].replace(".","").lower()
            
                if can_name_match and citation_author_last_name == author_items_last_name:
                    return True
                
    return False



def match_pub_authors_to_config_authors(authors_json, author_list):
    """Try to match authors in pub data to authors in config data.
    
    Goes through the author_list from a publication and tries matching to an author in 
    authors_json using firstname, lastname, and affiliations, or ORCID if ORCID
    is present.
    
    Args:
        authors_json (dict): keys are authors and values are author attributes. Matches authors JSON schema.
        author_list (list): list of dicts where each dict is the attributes of an author.
        
    Returns:
        author_list (list): either the author list with matched authors containing an additional author_id and/or ORCID attribute, or an empty list if no authors were matched.
    """
    publication_has_affiliated_author = False
    for author_items in author_list:
        is_collective_author = False
        has_collective_name = False
        if "collectivename" in author_items:
            is_collective_author = True
            if author_items["collectivename"] is not None:
                has_collective_name = True
                author_items_collective_name = author_items["collectivename"]
        
        else:
            can_name_match = True
            if author_items["firstname"] is None:
                can_name_match = False
            else:
                author_items_first_name = author_items["firstname"].replace(".","").lower()
            
            if author_items["lastname"] is None:
                can_name_match = False
            else:
                author_items_last_name = author_items["lastname"].replace(".","").lower()
            
            if author_items["affiliation"] is None:
                can_name_match = False
            else:
                author_items_affiliation = author_items["affiliation"].lower()
                        
        for author, author_attributes in authors_json.items():
            ## Try to match on ORCID.
            if author_items["ORCID"] and "ORCID" in author_attributes and author_items["ORCID"] == author_attributes["ORCID"]:
                publication_has_affiliated_author = True
                author_items["author_id"] = author
                break
            
            author_json_is_collective_author = False
            author_json_has_collective_name = False
            if "collective_name" in author_attributes:
                author_json_is_collective_author = True
                if author_attributes["collective_name"] is not None:
                    author_json_has_collective_name = True
                    author_json_collective_name = author_attributes["collective_name"]
            
            ## If it is a collective author then match on collective_name, else match on first and last.
            if is_collective_author:
                if not has_collective_name or \
                   not author_json_is_collective_author or \
                   not author_json_has_collective_name:
                    continue
                                
                if do_strings_fuzzy_match(author_json_collective_name, author_items_collective_name):
                       publication_has_affiliated_author = True
                       author_items["author_id"] = author
                       author_items["ORCID"] = author_attributes["ORCID"] if "ORCID" in author_attributes and author_items["ORCID"] is None else author_items["ORCID"]
                       break
            
            elif not author_json_is_collective_author:
                author_json_first_name = author_attributes["first_name"].replace(".","").lower()
                author_json_last_name = author_attributes["last_name"].replace(".","").lower()
            
                ## Match the author's first and last name and then match affiliations.
                ## The first name is matched with an additional .* to try and allow for the addition of initials, but this could cause bad matches. 
                ## For example the name Hu will match Hubert. Counting on the last name to reduce errors.
                ## Note that the PubMed query will return publications where the author is just a collaborater, so not finding a match in the authors isn't uncommon.
                if can_name_match and re.match(_generate_first_name_match_regex(author_json_first_name), author_items_first_name) and \
                   author_json_last_name == author_items_last_name:
                    ## affiliations in author_attributes are sets of strings so see if any are in the author_items string.
                    if any([affiliation.lower() in author_items_affiliation for affiliation in author_attributes["affiliations"]]):
                        publication_has_affiliated_author = True
                        author_items["author_id"] = author
                        author_items["ORCID"] = author_attributes["ORCID"] if "ORCID" in author_attributes and author_items["ORCID"] is None else author_items["ORCID"]
                        break
                
    if publication_has_affiliated_author:
        return author_list
    else:
        return []



def match_authors_in_prev_pub(prev_author_list, new_author_list):
    """Look for matching authors in previous pub data.
    
    Goes through the new_author_list and tries to find a match for each author 
    in the prev_author_list. Any authors that aren't matched are added to a 
    combined list. Both lists are expected to be a list of a dicts.
    
    {"firstname": author's first name,
     "lastname": author's last name,
     "author_id" : author's ID,
     "ORCID": ORCID ID}
    
    {"collectivename": collective name,
     "author_id": author's ID,
     "ORCID": ORCID ID}
    
    If author_id is missing or None in the prev_author_list, it will be updated 
    in the combined_author_list if the matched author in new_author_list has it. 
    The same can be said for the ORCID attribute.
    
    Args:
        prev_author_list (list): list of dicts where each dict is the attributes of an author.
        new_author_list (list): list of dicts where each dict is the attributes of an author.
        
    Returns:
        combined_author_list (list): the prev_author_list updated with "author_id" for dictionaries in the list that matched the given author.
    """
    combined_author_list = copy.deepcopy(prev_author_list)
    for new_author_attributes in new_author_list:
        is_collective_author = False
        has_collective_name = False
        if "collectivename" in new_author_attributes:
            is_collective_author = True
            if new_author_attributes["collectivename"] is not None:
                has_collective_name = True
                new_collective_name = new_author_attributes["collectivename"]
        
        else:
            can_name_match = True
            if new_author_attributes["firstname"] is None:
                can_name_match = False
            else:
                new_firstname_adjusted = new_author_attributes["firstname"].replace(".","").lower()
            
            if new_author_attributes["lastname"] is None:
                can_name_match = False
            else:
                new_lastname_adjusted = new_author_attributes["lastname"].replace(".","").lower()
        
        new_author_matched = False
        for i, prev_author_attributes in enumerate(prev_author_list):
            ## Try to match with author_id.
            if new_author_attributes.get("author_id", 0) == prev_author_attributes.get("author_id", 1):
                new_author_matched = True
                if prev_author_attributes["ORCID"] is None:
                    combined_author_list[i]["ORCID"] = new_author_attributes["ORCID"]
                break
            
            ## Try to match with ORCID.
            if new_author_attributes["ORCID"] and new_author_attributes["ORCID"] == prev_author_attributes["ORCID"]:
                new_author_matched = True
                if (author_id := new_author_attributes.get("author_id")) and not prev_author_attributes.get("author_id"):
                    combined_author_list[i]["author_id"] = author_id
                break
            
            prev_is_collective_author = False
            prev_has_collective_name = False
            if "collectivename" in prev_author_attributes:
                prev_is_collective_author = True
                if prev_author_attributes["collectivename"] is not None:
                    prev_has_collective_name = True
                    prev_collective_name = prev_author_attributes["collectivename"]
            
            ## If it is a collective author then match on collectivename, else match on first and last.
            if is_collective_author:
                if not has_collective_name or \
                   not prev_is_collective_author or \
                   not prev_has_collective_name:
                    continue
                
                if do_strings_fuzzy_match(prev_collective_name, new_collective_name):
                        new_author_matched = True
                        if (author_id := new_author_attributes.get("author_id")) and not prev_author_attributes.get("author_id"):
                            combined_author_list[i]["author_id"] = author_id
                        if prev_author_attributes["ORCID"] is None:
                            combined_author_list[i]["ORCID"] = new_author_attributes["ORCID"]
                        break
            
            elif not prev_is_collective_author:
                ## If the first and last names are missing for either author then they can't be matched.
                if prev_author_attributes["firstname"] is None or \
                   prev_author_attributes["lastname"] is None or \
                   not can_name_match:
                    continue
                    
                prev_firstname_adjusted = prev_author_attributes["firstname"].replace(".","").lower()
                prev_lastname_adjusted = prev_author_attributes["lastname"].replace(".","").lower()
                
                if re.match(_generate_first_name_match_regex(new_firstname_adjusted), prev_firstname_adjusted) and \
                   new_lastname_adjusted == prev_lastname_adjusted:
                        new_author_matched = True
                        if (author_id := new_author_attributes.get("author_id")) and not prev_author_attributes.get("author_id"):
                            combined_author_list[i]["author_id"] = author_id
                        if prev_author_attributes["ORCID"] is None:
                            combined_author_list[i]["ORCID"] = new_author_attributes["ORCID"]
                        break
        
        if not new_author_matched:
            combined_author_list.append(new_author_attributes)
                
    return combined_author_list



def _match_references_in_prev_pub(prev_reference_list, new_reference_list):
    """Look for matching references in previous pub data.
    
    Goes through the new_reference_list and tries to find a match for each reference 
    in the prev_reference_list. Any references that aren't matched are added to a 
    combined list. Both lists are expected to be a list of a dicts.
    
    {"citation": reference citation,
     "doi": reference DOI,
     "pubmed_id" : reference PMID,
     "PMCID": reference PMCID}
    
    Fields that are missing or None in the prev_reference_list will be updated 
    in the combined_reference_list if the matched reference in new_reference_list has it.
    
    Args:
        prev_reference_list (list): list of dicts where each dict is the attributes of a reference.
        new_reference_list (list): list of dicts where each dict is the attributes of a reference.
        
    Returns:
        combined_reference_list (list): the prev_reference_list updated with keys for dictionaries in the list that matched the given reference.
    """
    characters_to_remove = ['.', ',', ';', '(', ')', '[', ']', '{', '}']
    combined_reference_list = copy.deepcopy(prev_reference_list)
    matched_indexes = []
    for new_reference_attributes in new_reference_list:
        new_reference_matched = False
        for i, prev_reference_attributes in enumerate(prev_reference_list):
            ## There is a rare case where 2 different publications have the same title and 
            ## are referenced in the same paper. Need to disallow matching to the same reference 
            ## twice to avoid duplicates. Example paper: https://pubmed.ncbi.nlm.nih.gov/24404440/ 
            ## reference 1 and 3 have the same title, but are not the same reference.
            if i in matched_indexes:
                continue
            new_doi = new_reference_attributes.get("doi")
            prev_doi = prev_reference_attributes.get("doi")
            new_pmid = new_reference_attributes.get("pubmed_id")
            prev_pmid = prev_reference_attributes.get("pubmed_id")
            new_pmcid = new_reference_attributes.get("PMCID")
            prev_pmcid = prev_reference_attributes.get("PMCID")
            new_citation = new_reference_attributes.get("citation")
            prev_citation = prev_reference_attributes.get("citation")
            new_title = new_reference_attributes.get("title")
            prev_title = prev_reference_attributes.get("title")
                        
            if (new_doi and prev_doi and new_doi.lower() == prev_doi.lower()) or\
               (new_pmid and prev_pmid and new_pmid == prev_pmid) or\
               (new_pmcid and prev_pmcid and new_pmcid == prev_pmcid) or\
               (new_title and prev_title and \
               (do_strings_fuzzy_match(new_title, prev_title))) or\
               (new_title and prev_citation and new_title.lower() in prev_citation.lower()) or\
               (new_citation and prev_title and prev_title.lower() in new_citation.lower()) or\
               ((percentages := _compute_common_phrase_percent(prev_citation, new_citation, characters_to_remove, 4)) and\
                (percentages[0] >= 85 or percentages[1] >= 85)):
                _update(combined_reference_list[i], new_reference_attributes)
                new_reference_matched = True
                matched_indexes.append(i)
                break
            
        if not new_reference_matched:
            combined_reference_list.append(new_reference_attributes)
                
    return combined_reference_list



def _compute_common_phrase_percent(prev_citation, new_citation, characters_to_remove, min_len=2):
    """Find common phrases between prev_citation and new_citation and return percentage.
    
    Determine the common subphrases between the citations and then divide the length of the 
    common subphrases with the length of the common subphrases plus the uncommon subphrases 
    left for each citation and multiply by 100 to get the percentage of characters that are 
    common between the common+uncommon and common subphrases. If either citation is None, then 
    return None.
    
    Example:
        prev_citation = 'open js foundation 2019  accessed on 1 january 2023  available online:  https://openjsforg/'
        new_citation = '2023 january 01 open js foundation available online: https://openjsforg/'
        
        common_base_string = ' https://openjsforg/open js foundation available online:  january 2023 '
        
        prev_common_plus_uncommon = ' https://openjsforg/open js foundation available online:  january 2023 2019  accessed on 1'
        new_common_plus_uncommon = ' https://openjsforg/open js foundation available online:  january 2023 01'
        
        prev_percent = 78.88888888888889
        new_percent = 97.26027397260275
    
    Args:
        prev_citation (str|None): a string to find common subphrases with another string.
        new_citation (str|None): a string to find common subphrases with another string.
        characters_to_remove (list): a list of characters or strings to remove from each citation before finding common subphrases.
        min_len (int): the minimum length of a subphrase.
    
    Returns:
        ((int, int)|None): if either citation is None or empty after character removal and stripping, then return None, else the percentage of common to uncommon phrase length for each citation.
    """
    if prev_citation and new_citation:
        citation_strip_regex = "|".join([f"\\{char}" for char in characters_to_remove])
        # citation_strip_regex = r"\.|,|;|\(|\)|\[|\]|\{|\}"
        stripped_prev_citation = re.sub(citation_strip_regex, "", prev_citation.lower())
        stripped_new_citation = re.sub(citation_strip_regex, "", new_citation.lower())
        
        common_subphrases = find_common_subphrases(stripped_prev_citation, stripped_new_citation, min_len)
        
        prev_citation_common_phrases_removed = stripped_prev_citation
        new_citation_common_phrases_removed = stripped_new_citation
        for phrase in common_subphrases:
            prev_citation_common_phrases_removed = prev_citation_common_phrases_removed.replace(phrase.strip(), "")
            new_citation_common_phrases_removed = new_citation_common_phrases_removed.replace(phrase.strip(), "")
        common_base_string = "".join(common_subphrases)
        prev_common_denom = len(common_base_string + prev_citation_common_phrases_removed.strip())
        new_common_denom = len(common_base_string + new_citation_common_phrases_removed.strip())
        if prev_common_denom == 0 or new_common_denom == 0: 
            return None
        prev_common_percentage = len(common_base_string) / prev_common_denom * 100
        new_common_percentage = len(common_base_string) / new_common_denom * 100
        
        return prev_common_percentage, new_common_percentage
    else:
        return None



def create_pub_dict_for_saving_PubMed(pub, include_xml=False):
    """Convert pymed.PubMedArticle to a dictionary and modify it for saving.
    
    Converts a pymed.PubMedArticle to a dictionary, deletes the "xml" key if include_xml is False, and 
    converts the "publication_date" key to a string.
    
    Args:
        pub (pymed.PubMedArticle): publication to convert to a dictionary.
        include_xml (bool): if True, include the raw XML query in the key "xml".
        
    Returns:
        pub_id (str): the ID of the publication (DOI or PMID).
        pub_dict (dict): pub converted to a dictionary. Keys are "pubmed_id", "title", "abstract", 
        "keywords", "journal", "publication_date", "authors", "methods", "conclusions", "results", "copyrights", and "doi"
    """
    
    pub_dict = pub.toDict()
        
    if (pmid := pub_dict["xml"].find("PubmedData/ArticleIdList/ArticleId[@IdType='pubmed']")) is not None:
        pub_dict["pubmed_id"] = pmid.text
    else:
        pub_dict["pubmed_id"] = None
    
    if (doi := pub_dict["xml"].find("PubmedData/ArticleIdList/ArticleId[@IdType='doi']")) is not None:
        pub_dict["doi"] = doi.text.lower()
    else:
        pub_dict["doi"] = None
        
    if (pmc := pub_dict["xml"].find("PubmedData/ArticleIdList/ArticleId[@IdType='pmc']")) is not None:
        pub_dict["PMCID"] = pmc.text
    else:
        pub_dict["PMCID"] = None
    
    pub_dict["grants"] = [grant.text for grant in pub.xml.findall(".//GrantID")]
    
    references = []
    for reference in pub_dict["xml"].findall(".//Reference"):
        if (text := reference.find("Citation")) is not None:
            citation = text.text
        else:
            citation = None
        
        if (text := reference.find("ArticleIdList/ArticleId[@IdType='pmc']")) is not None:
            ref_pmc = text.text
        else:
            ref_pmc = None
        
        if (text := reference.find("ArticleIdList/ArticleId[@IdType='pubmed']")) is not None:
            ref_pmid = text.text
        else:
            ref_pmid = None
        
        if (text := reference.find("ArticleIdList/ArticleId[@IdType='doi']")) is not None:
            ref_doi = text.text.lower()
        else:
            ref_doi = None
        
        temp_dict = {"citation":citation,
                     "title": None,
                     "PMCID":ref_pmc,
                     "pubmed_id":ref_pmid,
                     "doi":ref_doi}
        
        ## Only add references that have at least 1 non-null value.
        if not all([value is None for value in temp_dict.values()]):
            references.append(temp_dict)
    pub_dict["references"] = references
    
    authors = []
    for author in pub_dict["xml"].findall(".//Author"):
        last_name = None
        if (text := author.find("LastName")) is not None:
            last_name = text.text
        
        collective_name = None
        if (text := author.find("CollectiveName")) is not None:
            collective_name = text.text
        
        first_name = None
        if (text := author.find("ForeName")) is not None:
            first_name = text.text
        
        initials = None
        if (text := author.find("Initials")) is not None:
            initials = text.text
        
        affiliation = None
        if (affiliations := author.findall("AffiliationInfo/Affiliation")) is not None:
            affiliation = '\n'.join([text.text for text in affiliations]) if affiliations else None
        
        orcid = None
        if (text := author.find("Identifier[@Source='ORCID']")) is not None:
            orcid = extract_ORCID_from_string(text.text)
        
        if collective_name:
            temp_dict = {"collectivename":collective_name,
                         "ORCID":orcid,
                         "author_id":None}
        else:
            temp_dict = {"lastname":last_name if last_name else collective_name,
                         "firstname": first_name,
                         "initials":initials,
                         "affiliation":affiliation,
                         "ORCID":orcid,
                         "author_id":None}
        
        ## Only add authors that have at least 1 non-null value.
        if not all([value is None for value in temp_dict.values()]):
            authors.append(temp_dict)
    pub_dict["authors"] = authors
        
    if not include_xml:
        del pub_dict["xml"]
    else:
        pub_dict["xml"] = str(ET.tostring(pub_dict["xml"]), encoding="utf-8")
    
    if pub_dict["publication_date"]:
        pub_dict["publication_date"] = {"year":pub_dict["publication_date"].year, "month":pub_dict["publication_date"].month, "day":pub_dict["publication_date"].day}
    else:
        pub_dict["publication_date"] = {"year":None, "month":None, "day":None}
    
    pub_id = DOI_URL + pub_dict["doi"] if pub_dict["doi"] else pub_dict["pubmed_id"]
    
    return pub_id, pub_dict



def create_pub_dict_for_saving_Crossref(work, prev_query):
    """Create the standard pub_dict from a Crossref query work dict.
    
    Args:
        work (dict): the dictionary for a publication returned in a Crossref query.
        prev_query (dict|None): a dictionary containing publications from a previous query, used for message printing.
    
    Returns:
        pub_id (str|None): the ID of the publication (DOI, PMID, or URL). If None, an ID couldn't be determined.
        pub_dict (dict|None): the standard pub_dict with values filled in from the Crossref publication. If None, an ID couldn't be determined.
    """
    if "title" in work:
        title = work["title"][0]
    else:
        return None, None
    
    ## Look for DOI
    doi = work["DOI"].lower() if "DOI" in work else None
    
    ## Look for external URL
    if "URL" in work:
        external_url = work["URL"]
    elif "link" in work:
        external_url = [link["URL"] for link in work["link"] if "URL" in link and link["URL"]][0]
        if not external_url:
            external_url = None
    else:
        external_url = None
                    
    
    if doi:
        pub_id = DOI_URL + doi
    elif external_url:
        pub_id = external_url
    elif not prev_query:
        vprint("Warning: Could not find a DOI or external URL for a publication when searching Crossref. It will not be in the publications.", verbosity=1)
        vprint("Title: " + title, verbosity=1)
        return None, None
    else:
        return None, None
    
    ## Determine authors and put them in unified form.
    new_author_list = []
    if "author" in work:
        for cr_author_dict in work["author"]:
            temp_dict = {}
            
            orcid = None
            if "ORCID" in cr_author_dict:
                orcid = extract_ORCID_from_string(cr_author_dict["ORCID"])
            temp_dict["ORCID"] = orcid
            
            temp_dict["author_id"] = None
            
            ## If "name" is in dict then it is a collective author and not an individual.
            if "name" in cr_author_dict:
                temp_dict["collectivename"] = cr_author_dict["name"]
            
            else:
                temp_dict["lastname"] = cr_author_dict.get("family")
                temp_dict["firstname"] = cr_author_dict.get("given")
                temp_dict["initials"] = None
                
                affiliations = []
                if cr_author_dict["affiliation"]:
                    for affiliation in cr_author_dict["affiliation"]:
                        if aff_text := affiliation.get("name"):
                            affiliations.append(aff_text)
                temp_dict["affiliation"] = '\n'.join(affiliations) if affiliations else None
            
            ## Only add authors that have at least 1 non-null value.
            if not all([value is None for value in temp_dict.values()]):
                new_author_list.append(temp_dict)
    
    
    ## Find published date
    date_found = False
    if "published" in work:
        date_key = "published"
        date_found = True
    elif "published-online" in work:
        date_key = "published-online"
        date_found = True
    elif "published-print" in work:
        date_key = "published-print"
        date_found = True
    
    if date_found:
        date_length = len(work[date_key]["date-parts"][0])
        
        if date_length > 2:
            publication_year = work[date_key]["date-parts"][0][0]
            publication_month = work[date_key]["date-parts"][0][1]
            publication_day = work[date_key]["date-parts"][0][2]
        elif date_length > 1:
            publication_year = work[date_key]["date-parts"][0][0]
            publication_month = work[date_key]["date-parts"][0][1]
            publication_day = None
        elif date_length > 0:
            publication_year = work[date_key]["date-parts"][0][0]
            publication_month = None
            publication_day = None
        else:
            publication_year = None
            publication_month = None
            publication_day = None
    else:
        publication_year = None
        publication_month = None
        publication_day = None
    
    
    ## look for grants in results
    found_grants = []
    if "funder" in work:
        for funder in work["funder"]:
            if awards := funder.get("award"):
                found_grants += awards
        
    ## Look for journal
    journal = work["publisher"] if "publisher" in work else None
    
    ## Look for references and put them in unified form.
    references = []
    if "reference" in work:
        for reference in work["reference"]:
            if "DOI" in reference:
                ref_doi = reference["DOI"]
            else:
                ref_doi = None
            
            if ref_title := reference.get("article-title"):
                pass
            else:
                ref_title = reference.get("series-title")
            
            reference_dict = {"citation":reference.get("unstructured"),
                              "title":ref_title,
                              "PMCID":None,
                              "pubmed_id":None,
                              "doi":ref_doi}
            if all([value is None for value in reference_dict.values()]):
                continue
            else:
                references.append(reference_dict)
        

    ## Build the pub_dict from what we were able to collect.
    pub_dict = copy.deepcopy(PUBLICATION_TEMPLATE)
    
    pub_dict["doi"] = doi
    pub_dict["publication_date"]["year"] = publication_year
    pub_dict["publication_date"]["month"] = publication_month
    pub_dict["publication_date"]["day"] = publication_day
    pub_dict["journal"] = journal
    pub_dict["grants"] = found_grants
    pub_dict["authors"] = new_author_list
    pub_dict["title"] = title
    pub_dict["references"] = references
    
    return pub_id, pub_dict



ORCID_regex = r"(\d{4}-\d{4}-\d{4}-\d{3}[0,1,2,3,4,5,6,7,8,9,X])"
alt_ORCID_regex = r"(\d{4}\d{4}\d{4}\d{3}[0,1,2,3,4,5,6,7,8,9,X])"
def extract_ORCID_from_string(string):
    """Extract an ORCID ID from a string.
    
    Args:
        string (str): the string to extract the ID from.
        
    Returns:
        (str|None): either the extracted ID as a string or None.
    """
    
    if re_match := re.search(ORCID_regex, string):
        return re_match.groups()[0]
    elif re_match := re.search(alt_ORCID_regex, string):
        captured_string = re_match.groups()[0]
        new_string = captured_string[0:4] + '-' + captured_string[4:8] + '-' + captured_string[8:12] + '-' + captured_string[12:]
        return new_string
    else:
        return None



def is_fuzzy_match_to_list(str_to_match, list_to_match):
    """True if string is a 90 or higher ratio match to any string in list, False otherwise.
    
    Args:
        str_to_match (str): string to compare with list
        list_to_match (list): list of strings to compare with str_to_match
        
    Returns:
        (bool): True if str_to_match is a match to any string in list_tp_match, False otherwise.
    """
    # str_to_match = str_to_match.lower()
    # list_to_match = [list_string.lower() for list_string in list_to_match]
    
    return any([do_strings_fuzzy_match(str_to_match, list_string) for list_string in list_to_match])



def fuzzy_matches_to_list(str_to_match, list_to_match):
    """Return strings and indexes for strings with match ratio that is 90 or higher.
    
    Args:
        str_to_match (str): string to compare with list
        list_to_match (list): list of strings to compare with str_to_match
        
    Returns:
        (list): list of matches (tuples) with each element being the string and its index in list_to_match. [(9, "title 1"), ...]
    """
    # str_to_match = str_to_match.lower()
    # list_to_match = [list_string.lower() for list_string in list_to_match]
    
    return [(index, list_string) for index, list_string in enumerate(list_to_match) if do_strings_fuzzy_match(str_to_match, list_string)]



def is_pub_in_publication_dict(pub_id, title, publication_dict, titles=None):
    """True if pub_id is in publication_dict or title is a fuzzy match to titles in titles.
    
    Check whether the pub_id is in publication_dict. If it isn't then see if there is 
    a fuzzy match in titles. If titles is not provided then get a list of titles 
    from publication_dict.
    
    Args:
        pub_id (str): pub_id to check against in publication_dict to see if it already exists.
        title (str): title corresponding to pub_id to check against titles in publication_dict.
        publication_dict (dict): keys are pub_ids and values are pub attributes.
        titles (list|None): list of strings that should be titles to fuzzy match to title.
        
    Returns:
        (bool): True if the pub_id is in publication_dict or title is fuzzy matched in titles, False otherwise
    """
    
    if pub_id.lower() in publication_dict:
        return True
    
    if not titles:
        titles = [pub_attr["title"] for pub_attr in publication_dict.values() if pub_attr["title"]]
    
    return True if is_fuzzy_match_to_list(title, titles) else False



def get_pub_id_in_publication_dict(pub_id, title, publication_dict):
    """Get the pub_id in publication_dict for the publication that matches the given pub_id or fuzzy matches a title.
    
    Check whether the pub_id is in publication_dict. If it isn't then see if there is 
    a fuzzy match in titles. It is assumed every dictionary in publication_dict will 
    have a "title" key with a string value.
    
    Args:
        pub_id (str): pub_id to check against in publication_dict to see if it already exists.
        title (str): title corresponding to pub_id to check against titles in publication_dict.
        publication_dict (dict): keys are pub_ids and values are pub attributes.
        
    Returns:
        (str|None): the pub_id matched in publication_dict or None if nothing was found.
    """
    
    if pub_id.lower() in publication_dict:
        return pub_id.lower()
    
    for key, value in publication_dict.items():
        if do_strings_fuzzy_match(title, value["title"]):
            return key
    
    return None



def find_duplicate_citations(tokenized_citations):
    """Find citations that are duplicates of each other in tokenized_citations.
    
    Citations can be duplicates in 3 ways. Same PMID, same DOI, or similar enough titles.
    The function goes through each citation and looks for matches on these criteria.
    Then the matches are compared to create unique sets. For instance if citation 1
    matches the PMID in citation 2, and citation 2 matches the DOI in citation 3, but 
    citation 1 and 3 don't match a duplicate set containing all 3 is created. The 
    unique duplicate sets are returned as a list of sorted lists.
    
    Args:
        tokenized_citation (list): list of dictionaries where each dictionary is a citation. Matches the tokenized_reference.json schema.
        
    Returns:
        unique_duplicate_sets (list): list of lists where each element is a list of indexes in tokenized_citations that match each other. The list of indexes is sorted in ascending order.
    """
    
    titles_list = [citation["title"] for citation in tokenized_citations if citation["title"]]
    
    pmids = {}
    dois = {}
    titles = {}
    for count, citation in enumerate(tokenized_citations):
        
        if citation["PMID"]:
            if citation["PMID"] in pmids:
                pmids[citation["PMID"]].append(count)
            else:
                pmids[citation["PMID"]] = [count]
                
        if citation["DOI"]:
            doi = citation["DOI"].lower()
            if doi in dois:
                dois[doi].append(count)
            else:
                dois[doi] = [count]
                
        if citation["title"]:
            fuzzy_matches = fuzzy_matches_to_list(citation["title"], titles_list)
            if len(fuzzy_matches) > 1 and fuzzy_matches[0][1] in titles:
                titles[fuzzy_matches[0][1]].append(count)
            else:
                titles[citation["title"]] = [count]
                
    
    matching_pmids = {tuple(indexes) for pmid, indexes in pmids.items() if len(indexes) > 1}
    matching_dois = {tuple(indexes) for doi, indexes in dois.items() if len(indexes) > 1}
    matching_titles = {tuple(indexes) for title, indexes in titles.items() if len(indexes) > 1}
    
    all_matches = matching_pmids | matching_dois | matching_titles
    
    ## Matches need to be matched recursively to find true sets of matches.
    ## For example if 1 and 2 are found to match on title and 2 and 3 are found to match 
    ## on PMID then we need 1 set of (1,2,3) instead of 2 sets of (1,2) (2,3).
    duplicates_dict = {}
    for index_tuple in list(all_matches):
        for index in index_tuple:
            if index in duplicates_dict:
                for index2 in index_tuple:
                    if index2 != index:
                        duplicates_dict[index].add(index2)
            else:
                duplicates_dict[index] = set()
                for index2 in index_tuple:
                    if index2 != index:
                        duplicates_dict[index].add(index2)
                        
    for index, matches in duplicates_dict.items():
        temp_set = matches
        set_before_loop = matches
        while(True):
            for match in matches:
                temp_set = temp_set | duplicates_dict[match]
            
            if set_before_loop == temp_set:
                break
            else:
                set_before_loop = temp_set
                
        duplicates_dict[index] = temp_set
        
    unique_duplicate_sets = {tuple(sorted(duplicate_set)) for duplicate_set in duplicates_dict.values()}
    unique_duplicate_sets = [list(duplicate_set) for duplicate_set in unique_duplicate_sets]
    for duplicate_set in unique_duplicate_sets:
        duplicate_set.sort()
    
    return unique_duplicate_sets
    
    

def are_citations_in_pub_dict(tokenized_citations, pub_dict):
    """Determine which citations in tokenized_citations are in pub_dict.
    
    For each citation in tokenized_citations see if it is in pub_dict. Will be 
    True for a citation if the PMID matches, DOI matches, or the title is similar 
    enough.
    
    Args:
        tokenized_citation (list): list of dictionaries where each dictionary is a citation. Matches the tokenized_reference.json schema.
        pub_dict (dict): schema matches the publication.json schema.
    
    Returns:
        (list): list of bools, True if the citation at that index is in pub_dict, False otherwise.
    """
    
    pub_titles = [pub["title"] for pub in pub_dict.values() if pub["title"]]
    pub_dois = [doi.lower() for pub in pub_dict.values() if (doi := normalize_DOI(pub["doi"]))]
    pub_pmids = [pub["pubmed_id"] for pub in pub_dict.values() if pub["pubmed_id"]]
    
    return [True if citation["PMID"] in pub_pmids or ((doi := normalize_DOI(citation["DOI"])) and doi.lower()) in pub_dois or is_fuzzy_match_to_list(citation["title"], pub_titles) else False for citation in tokenized_citations]


def normalize_DOI(doi_string):
    """
    """
    if doi_string:
        if doi_match := regex_match_return(r'https?://doi.org/(.*)', doi_string):
            return doi_match[0]
        return doi_string
    return None



# def _update_nulls(original_dict, upgrade_dict):
#     """Update a dictionary in a nested fashion.
    
#     Only updates original_dict if original_dict has a None value for a key. 
#     This is recursive, so if a dicitonary type is seen for the value then this 
#     is called on that nested dictionary.
    
#     Args:
#         original_dict (dict): the dictionary to update.
#         upgrade_dict (dict): the dictionary to update values from.
        
#     Returns:
#         original_dict (dict): the updated original_dict
#     """
#     for key, value in original_dict.items():
#         if isinstance(value, collections.abc.Mapping):
#             original_dict[key] = _update_nulls(original_dict[key], upgrade_dict[key])
#         elif value is None and upgrade_dict.get(key) is not None:
#             original_dict[key] = upgrade_dict[key]
    
#     return original_dict



def _update(original_dict, upgrade_dict):
    """Update a dictionary in a nested fashion.
    
    Only updates original_dict if original_dict has a None value for a key, or 
    if the key does not already exist in original_dict.
    This is recursive, so if a dicitonary type is seen for the value then this 
    is called on that nested dictionary.
    
    Args:
        original_dict (dict): the dictionary to update.
        upgrade_dict (dict): the dictionary to update values from.
        
    Returns:
        original_dict (dict): the updated original_dict
    """
    for key, value in upgrade_dict.items():
        if key not in original_dict:
            original_dict[key] = value
        elif isinstance(value, collections.abc.Mapping) and isinstance(original_dict[key], collections.abc.Mapping):
            original_dict[key] = _update(original_dict[key], upgrade_dict[key])
        elif value is not None and original_dict[key] is None:
            original_dict[key] = upgrade_dict[key]
    
    return original_dict



def _merge_pub_dicts(prev_dict, new_dict):
    """Merge information from 2 unified pub_dicts.
    
    Args:
        prev_dict (dict): the dictionary whose values will be updated.
        new_dict (dict): the dictionary whose values will be used to update prev_dict.
    
    Returns:
        prev_dict (dict): prev_dict with updated values
    """
    _update(prev_dict, new_dict)
    prev_dict["authors"] = match_authors_in_prev_pub(prev_dict["authors"], new_dict["authors"])
    prev_dict["references"] = _match_references_in_prev_pub(prev_dict["references"], new_dict["references"])
    new_grants = [grant for grant in new_dict["grants"] if grant not in prev_dict["grants"]]
    prev_dict["grants"] += new_grants
    
    return prev_dict



def find_common_subphrases(str1, str2, min_len=2):
    """Find all common subphrases between str1 and str2 longer than min_len.
    
    Modified from https://stackoverflow.com/a/63337541/19957088.
    Find all common subphrases between the 2 strings, but filer out common subphrases 
    between subphrases. For example, if "sand" is common between the 2 strings the 
    function will not return "and" unless there is another instance of "and" between 
    the 2 strings. A phrase is a string that must end in a space or be at the end 
    of the string. So "sand asdf" and "sand awer" will only match "sand " and not 
    "sand a". Spaces are expected to be meaningful. It is recommended to remove 
    punctuation from the strings.
    
    Args:
        str1 (str): one of the 2 strings to look for common substrings in.
        str2 (str): one of the 2 strings to look for common substrings in.
        min_len (int): if the length of a substring is less than this, then ignore it.
        
    Returns:
        cs_array (list): a list of the common substrings.
    """
    len1, len2 = len(str1), len(str2)

    if len1 > len2:
        str1, str2 = str2, str1 
        len1, len2 = len2, len1
    
    cs_array=[]
    cs_array_index = []
    for i in range(len1, min_len-1, -1):
        for k in range(len1-i+1):
            end_index = i + k
            sub_string = str1[k:end_index]
            if sub_string[-1] != " " and end_index < len1 and sub_string[0] != " " and k > 0:
                continue
            if sub_string in str2:
                flag = 1
                for m in range(len(cs_array)):
                    if (sub_string in cs_array[m] and\
                       ## Have to make sure the sub_string is in the same index range as what it matched.
                       ## Otherwise, "and" won't get matched between 'sand asdf and' and 'sand and'.
                       k >= cs_array_index[m][0] and end_index <= cs_array_index[m][1]) or (\
                       ## We don't want to return overlapping substrings, without the below check 'sand asdf and'
                       ## and 'sand and' will return 'sand a' and 'and' instead of 'sand a' and 'nd'.
                       ## This is undesirable for us because we plan to remove the found substrings 
                       ## after this function, and overlapping substrings make that harder.
                       (k >= cs_array_index[m][0] and k < cs_array_index[m][1])):
                        flag=0
                        break
                if flag == 1:
                    cs_array.append(sub_string)
                    cs_array_index.append([k, end_index])
    return cs_array

## Good test phrases to understand how the function works.
# find_common_subphrases('sandc cand','sandb asdf and', 2)
# find_common_subphrases('sand and','sand asdf and', 2)
# find_common_subphrases('sand and','sand qsdf and', 2)




# -*- coding: utf-8 -*-
"""
Reference Search Webio
~~~~~~~~~~~~~~~~~~~~~~

Internet interfacing for reference_search.
"""

import time
import copy
import sys
import os
import re

import pymed
import scholarly
import habanero
import bs4

from . import helper_functions
from . import citation_parsing
from . import webio
from . import fileio
from . import user_input_checking


TOOL = webio.TOOL
DOI_URL = webio.DOI_URL

PUBLICATION_TEMPLATE = webio.PUBLICATION_TEMPLATE


def build_pub_dict_from_PMID(PMID_list, from_email):
    """Query PubMed for each PMID and build a dictionary of the returned data.
    
    Args:
        PMID_list (list): A list of PMIDs as strings.
        from_email (str): An email address to use when querying PubMed.
        
    Returns:
        publication_dict (dict): keys are pulication ids and values are a dictionary with publication attributes.
    """
    
    pubmed = pymed.PubMed(tool=TOOL, email=from_email)
    
    publication_dict = dict()
    
    for PMID_to_search in PMID_list:
        
        publications = pubmed.query(PMID_to_search, max_results=10)
        
        for pub in publications:
            pub_id, pub_dict = helper_functions.create_pub_dict_for_saving_PubMed(pub)
            
            if pub_dict["pubmed_id"] == PMID_to_search:
                publication_dict[pub_id] = pub_dict
                break
                
        time.sleep(1)

    return publication_dict



# def search_references_on_PubMed(running_pubs, tokenized_citations, from_email, prev_query=None):
#     """Searhes PubMed for publications matching the citations.
    
#     For each citation in tokenized_citations PubMed is queried for the publication. 
#     If the publication is already in running_pubs then missing information will be 
#     filled in if possible.
    
#     Args:
#         running_pubs (dict): dictionary of publications matching the JSON schema for publications.
#         tokenized_citations (list): list of citations parsed from a source. Each citation is a dict {"authors", "title", "DOI", "PMID", "reference_line", "pub_dict_key"}.
#         from_email (str): used in the query to PubMed.
#         prev_query (list|None): a list of lists containing publications from a previous call to this function. [[pub1, ...], [pub1, ...], ...]
        
#     Returns:
#         running_pubs (dict): keys are pulication ids and values are a dictionary with publication attributes
#         matching_key_for_citation (list): list of keys to the publication matching the citation at the same index
#         all_pubs (list): list of lists, each index is the pubs searched through after querying until the citation was matched
#     """
       
#     # initiate PubMed API
#     pubmed = pymed.PubMed(tool=TOOL, email=from_email)
    
#     all_pubs = []
#     matching_key_for_citation = []
    
#     for i, citation in enumerate(tokenized_citations):
#         all_pubs.append([])
        
#         if citation["PMID"]:
#             query_string = citation["PMID"]
#         elif citation["DOI"]:
#             query_string = citation["DOI"]
#         elif citation["title"]:
#             query_string = citation["title"]
#         else:
#             matching_key_for_citation.append(None)
#             continue
#         publications = pubmed.query(query_string, max_results=10) if not prev_query else prev_query[i]
        
#         citation_matched_to_pub = False
#         for pub in publications:
#             if not isinstance(pub, pymed.article.PubMedArticle):
#                 continue
#             all_pubs[i].append(pub)
            
#             pub_dict = helper_functions.create_pub_dict_for_saving_PubMed(pub)
#             pub_id = pub_dict["doi"] if pub_dict["doi"] else pub_dict["pubmed_id"]
            
#             ## Match publication to the citation.
#             if citation["PMID"] == pub_dict["pubmed_id"]:
#                 citation_matched_to_pub = True
#             elif pub.doi and citation["DOI"] and citation["DOI"].lower() == pub.doi.lower():
#                 citation_matched_to_pub = True
#             else:
#                 has_matching_author = helper_functions.match_pub_authors_to_citation_authors(citation["authors"], pub_dict["authors"])
#                 if has_matching_author and helper_functions.do_strings_fuzzy_match(citation["title"], pub_dict["title"]):
#                     citation_matched_to_pub = True
                               
#             if matching_pub_id := helper_functions.get_pub_id_in_publication_dict(pub_id, pub.title, running_pubs):
#                 if "PubMed" in running_pubs[matching_pub_id]["queried_sources"]:
#                     if not citation_matched_to_pub:
#                         continue
#                     matching_key_for_citation.append(matching_pub_id)
#                     break
                
#                 helper_functions._merge_pub_dicts(running_pubs[matching_pub_id], pub_dict)
#                 running_pubs[matching_pub_id]["queried_sources"].append("PubMed")
#                 if citation_matched_to_pub:
#                     matching_key_for_citation.append(matching_pub_id)
#                     break
            
#             else:
#                 if not citation_matched_to_pub:
#                     continue
#                 pub_dict["queried_sources"] = ["PubMed"]
#                 running_pubs[pub_id] = pub_dict
#                 matching_key_for_citation.append(pub_id)
#                 break
                            
#         if not citation_matched_to_pub:
#             matching_key_for_citation.append(None)
#         time.sleep(1)
        
#     return running_pubs, matching_key_for_citation, all_pubs




# def search_references_on_Crossref(running_pubs, tokenized_citations, mailto_email, prev_query=None):
#     """Searhes Crossref for publications matching the citations.
    
#     For each citation in tokenized_citations Crossref is queried for the publication. 
#     If the publication is already in running_pubs then missing information will be 
#     filled in if possible.
    
#     Args:
#         running_pubs (dict): dictionary of publications matching the JSON schema for publications.
#         tokenized_citations (list): list of citations parsed from a source. Each citation is a dict {"authors", "title", "DOI", "PMID", "reference_line", "pub_dict_key"}.
#         mailto_email (str): used in the query to Crossref.
#         prev_query (list|None): a list of lists containing publications from a previous call to this function. [[pub1, ...], [pub1, ...], ...]
        
#     Returns:
#         running_pubs (dict): keys are pulication ids and values are a dictionary with publication attributes
#         matching_key_for_citation (list): list of keys to the publication matching the citation at the same index
#         all_pubs (list): list of lists, each index is the pubs searched through after querying until the citation was matched
#     """
    
#     cr = habanero.Crossref(ua_string = "Academic Tracker (mailto:" + mailto_email + ")")
    
#     all_pubs = []
#     matching_key_for_citation = []
#     for i, citation in enumerate(tokenized_citations):
#         all_pubs.append([])
        
#         if not prev_query:
#             if citation["DOI"]:
#                 results = cr.works(ids = citation["DOI"])
#                 works = [results["message"]]
#             elif citation["title"]:
#                 results = cr.works(query_bibliographic = citation["title"], filter = {"type":"journal-article"}, limit = 10)
#                 works = results["message"]["items"]
#             else:
#                 matching_key_for_citation.append(None)
#                 continue
#         else:
#             works = prev_query[i]
                        
#         citation_matched_to_pub = False
#         for work in works:
#             all_pubs[i].append(work)
            
#             pub_id, pub_dict = helper_functions.create_pub_dict_for_saving_Crossref(work, prev_query)
            
#             if pub_id is None:
#                 continue
            
#             if citation["DOI"] == pub_dict["doi"]:
#                 citation_matched_to_pub = True
#             else:
#                 if "author" in work:
#                     has_matching_author = helper_functions.match_pub_authors_to_citation_authors(citation["authors"], pub_dict["authors"])
#                     if has_matching_author and helper_functions.do_strings_fuzzy_match(citation["title"], pub_dict["title"]):
#                         citation_matched_to_pub = True
                               
#             ## If the publication is already in running_pubs then try to update missing information.
#             if matching_pub_id := helper_functions.get_pub_id_in_publication_dict(pub_id, pub_dict["title"], running_pubs):
#                 if "Crossref" in running_pubs[matching_pub_id]["queried_sources"]:
#                     if not citation_matched_to_pub:
#                         continue
#                     matching_key_for_citation.append(matching_pub_id)
#                     break
                
#                 helper_functions._merge_pub_dicts(running_pubs[matching_pub_id], pub_dict)
#                 running_pubs[matching_pub_id]["queried_sources"].append("Crossref")
#                 if citation_matched_to_pub:
#                     matching_key_for_citation.append(matching_pub_id)
#                     break
            
#             else:
#                 if not citation_matched_to_pub:
#                     continue
#                 pub_dict["queried_sources"] = ["Crossref"]
#                 running_pubs[pub_id] = pub_dict
#                 matching_key_for_citation.append(pub_id)
#                 break
                    
#         if not citation_matched_to_pub:
#             matching_key_for_citation.append(None)
#         time.sleep(1)
            
#     return running_pubs, matching_key_for_citation, all_pubs
        



def search_references_on_source(source, running_pubs, tokenized_citations, mailto_email, prev_query=None):
    """Searhes source for publications matching the citations.
    
    For each citation in tokenized_citations the source is queried for the publication. 
    If the publication is already in running_pubs then missing information will be 
    filled in if possible.
    
    Possible sources are "Crossref" or "PubMed".
    
    Args:
        source (str): must be one of "Crossref" or "PubMed".
        running_pubs (dict): dictionary of publications matching the JSON schema for publications.
        tokenized_citations (list): list of citations parsed from a source. Each citation is a dict {"authors", "title", "DOI", "PMID", "reference_line", "pub_dict_key"}.
        mailto_email (str): email provided to the source when querying.
        prev_query (list|None): a list of lists containing publications from a previous call to this function. [[pub1, ...], [pub1, ...], ...]
        
    Returns:
        running_pubs (dict): keys are pulication ids and values are a dictionary with publication attributes
        matching_key_for_citation (list): list of keys to the publication matching the citation at the same index
        all_pubs (list): list of lists, each index is the pubs searched through after querying until the citation was matched
    """
       
    # initiate API
    if source == "PubMed":
        api = pymed.PubMed(tool=TOOL, email=mailto_email)
        query_function = _query_PubMed
        skip_pub_function = _pub_needs_skipped_PubMed
        pub_dict_creation_function = helper_functions.create_pub_dict_for_saving_PubMed
        pub_dict_creation_arguments = ["pub"]
    elif source == "Crossref":
        api = habanero.Crossref(ua_string = "Academic Tracker (mailto:" + mailto_email + ")")
        query_function = _query_Crossref
        skip_pub_function = _pub_needs_skipped_Crossref
        pub_dict_creation_function = helper_functions.create_pub_dict_for_saving_Crossref
        pub_dict_creation_arguments = ["pub", "prev_query"]
    else:
        helper_functions.vprint("Error: When searching references there was an attempt to query an unknown source, '" + source + "'.")
        sys.exit()
    
    all_pubs = []
    matching_key_for_citation = []
    
    for i, citation in enumerate(tokenized_citations):
        all_pubs.append([])
        
        if not prev_query:
            if not (publications := query_function(api, citation)):
                matching_key_for_citation.append(None)
                continue
        else:
            publications = prev_query[i]
        
        citation_matched_to_pub = False
        for pub in publications:
            if skip_pub_function(pub):
                continue
            all_pubs[i].append(pub)
            
            locals_ref = locals()
            pub_id, pub_dict = pub_dict_creation_function(*[locals_ref[arg] for arg in pub_dict_creation_arguments])
            
            if pub_id is None:
                continue
            
            ## Match publication to the citation.
            if pub_dict["pubmed_id"] and citation["PMID"] and pub_dict["pubmed_id"] == citation["PMID"]:
                citation_matched_to_pub = True
            elif pub_dict["doi"] and citation["DOI"] and citation["DOI"].lower() == pub_dict["doi"]:
                citation_matched_to_pub = True
            else:
                has_matching_author = helper_functions.match_pub_authors_to_citation_authors(citation["authors"], pub_dict["authors"])
                if has_matching_author and helper_functions.do_strings_fuzzy_match(citation["title"], pub_dict["title"]):
                    citation_matched_to_pub = True
            
                               
            if matching_pub_id := helper_functions.get_pub_id_in_publication_dict(pub_id, pub_dict["title"], running_pubs):
                if source in running_pubs[matching_pub_id]["queried_sources"]:
                    if not citation_matched_to_pub:
                        continue
                    matching_key_for_citation.append(matching_pub_id)
                    break
                
                helper_functions._merge_pub_dicts(running_pubs[matching_pub_id], pub_dict)
                running_pubs[matching_pub_id]["queried_sources"].append(source)
                if citation_matched_to_pub:
                    matching_key_for_citation.append(matching_pub_id)
                    break
            
            else:
                if not citation_matched_to_pub:
                    continue
                pub_dict["queried_sources"] = [source]
                running_pubs[pub_id] = pub_dict
                matching_key_for_citation.append(pub_id)
                break
                            
        if not citation_matched_to_pub:
            matching_key_for_citation.append(None)
        time.sleep(1)
        
    return running_pubs, matching_key_for_citation, all_pubs




def _query_PubMed(pubmed, citation):
    """Query PubMed with either the PMID, DOI, or title from citation.
    
    Args:
        pubmed (pymed.api.PubMed): api object from the pymed library.
        citation (dict): citation to query for.
    """
    
    if citation["PMID"]:
        query_string = citation["PMID"]
    elif citation["DOI"]:
        query_string = citation["DOI"]
    elif citation["title"]:
        query_string = citation["title"]
    else:
        return None
    return pubmed.query(query_string, max_results=10)



def _query_Crossref(cr, citation):
    """Query Crossref with either the PMID, DOI, or title from citation.
    
    Args:
        pubmed (habanero.crossref.crossref.Crossref): api object from the habanero library.
        citation (dict): citation to query for.
    """
    
    if citation["DOI"]:
        results = cr.works(ids = citation["DOI"])
        works = [results["message"]]
    elif citation["title"]:
        results = cr.works(query_bibliographic = citation["title"], filter = {"type":"journal-article"}, limit = 10)
        works = results["message"]["items"]
    else:
        return None
    return works
         

def _pub_needs_skipped_PubMed(pub):
    """Determine whether the queried pub from PubMed should be skipped or not.
    
    Args:
        pub (pymed.article.PubMedArticle|pymed.book.PubMedBookArticle): publication queried from PubMed book articles should be skipped.
    """
    
    return not isinstance(pub, pymed.article.PubMedArticle)


def _pub_needs_skipped_Crossref(pub):
    """Determine whether the queried pub from Crossref should be skipped or not.
    
    This is just here to work with the function, publications from Crossref should never be skipped.
    
    Args:
        pub (dict): publication queried from Crossref.
    """
    
    return False




def parse_myncbi_citations(url):
    """Tokenize the citations on a MyNCBI URL.
    
    Note that authors and title can be missing or empty from the webpage.
    This function assumes the url is the first page of the MyNCBI citations.
    The first page is tokenized and then each subsequent page is visited and 
    tokenized.
    
    Args:
        url (str): the url of the MyNCBI page.
        
    Returns:
        parsed_pubs (dict): the citations tokenized in a dictionary matching the tokenized citations JSON schema.    
    """
    
    ## Get the first page, find out the total pages, and parse it.
    url_str = webio.get_url_contents_as_str(url)
    if not url_str:
        helper_functions.vprint("Error: Could not access the MYNCBI webpage. Make sure the address is correct.")
        sys.exit()
    
    soup = bs4.BeautifulSoup(url_str, "html.parser")
    number_of_pages = int(soup.find("span", class_ = "totalPages").text)
    
    parsed_pubs = citation_parsing.tokenize_myncbi_citations(url_str)
    
    ## Parse the rest of the pages.
    new_url = url if url[-1] == "/" else url + "/"
    
    for i in range(2,number_of_pages+1):
        
        url_str = webio.get_url_contents_as_str(new_url + "?page=" + str(i))
        if not url_str:
            helper_functions.vprint("Error: Could not access page " + str(i) + " of the MYNCBI webpage. Aborting run.")
            sys.exit()
        
        temp_pubs = citation_parsing.tokenize_myncbi_citations(url_str)
        parsed_pubs += temp_pubs
        
    return parsed_pubs



def tokenize_reference_input(reference_input, MEDLINE_reference, remove_duplicates=True):
    """Tokenize the citations in reference_input.
    
    reference_input can be a URL or filepath. MyNCBI URLs are handled special, 
    but all other URLs are read as a text document and parsed line by line as 
    if they were a test document. If the format of the reference is MEDLINE then 
    set MEDLINE_reference to True and it will be parsed as such instead of line 
    by line. Citations are expected to be 1 per line otherwise.
    
    Args:
        reference_input (str): URL or filepath.
        MEDLINE_reference (bool): True if reference_input is in MEDLINE format.
        remove_duplicates (bool): if True, remove duplicate entries in tokenized citations.
        
    Returns:
        tokenized_citations (dict): the citations tokenized in a dictionary matching the tokenized citations JSON schema. 
    """
    
    ## Check the reference_input to see if it is json.
    extension = os.path.splitext(reference_input)[1][1:]
    if extension == "json":
        tokenized_citations = fileio.load_json(reference_input)
        user_input_checking.tok_reference_check(tokenized_citations)
    
    ## Check the reference file input and see if it is a URL
    elif re.match(r"http.*", reference_input):
        if re.match(r".*ncbi.nlm.nih.gov/myncbi.*", reference_input):
            tokenized_citations = parse_myncbi_citations(reference_input)
        else:
            document_string = webio.clean_tags_from_url(reference_input)
            
            if not document_string:
                helper_functions.vprint("Nothing was read from the URL. Make sure the address is correct.")
                sys.exit()
            
            tokenized_citations = citation_parsing.parse_text_for_citations(document_string)
    else:
        # Check the file extension and call the correct read in function.
        if extension == "docx":
            document_string = fileio.read_text_from_docx(reference_input)
        elif extension == "txt":
            document_string = fileio.read_text_from_txt(reference_input)
        else:
            helper_functions.vprint("Unknown file type for reference file.")
            sys.exit()
    
        if not document_string:
            helper_functions.vprint("Nothing was read from the reference file. Make sure the file is not empty or is a supported file type.")
            sys.exit()
        
        if MEDLINE_reference:
            tokenized_citations = citation_parsing.parse_MEDLINE_format(document_string)
        else:
            tokenized_citations = citation_parsing.parse_text_for_citations(document_string)
            
    if not tokenized_citations:
        helper_functions.vprint("Warning: Could not tokenize any citations in provided reference. Check setup and formatting and try again.")
        sys.exit()
    
    ## Look for duplicates in citations.
    if remove_duplicates:
        duplicate_citations = helper_functions.find_duplicate_citations(tokenized_citations)
        if duplicate_citations:
            helper_functions.vprint("Warning: The following citations in the reference file or URL appear to be duplicates based on identical DOI, PMID, or similar titles. They will only appear once in any outputs.", verbosity=1)
            helper_functions.vprint("Duplicates:", verbosity=1)
            for index_list in duplicate_citations:
                for index in index_list:
                    if tokenized_citations[index]["reference_line"]:
                        pretty_print = tokenized_citations[index]["reference_line"].split("\n")
                        pretty_print = " ".join([line.strip() for line in pretty_print])
                        helper_functions.vprint(pretty_print, verbosity=1)
                    else:
                        helper_functions.vprint(tokenized_citations[index]["title"], verbosity=1)
                    helper_functions.vprint("", verbosity=1)
                helper_functions.vprint("\n", verbosity=1)
            
            indexes_to_remove = [index for duplicate_set in duplicate_citations for index in duplicate_set[1:]]
            
            tokenized_citations = [citation for count, citation in enumerate(tokenized_citations) if not count in indexes_to_remove]
        
    return tokenized_citations






